LICENSE = [[
mod_classic.lua

Copyright (c) 2013 Phil Hassey

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

]]

BETTER_LICENCE = [[
mod_esparano_frustration.lua

All hail esparano.  By using this mod, you acknowledge to the best of your knowledge that ESPARANO IS THE BEST GALCON PLAYER OF ALL TIME, INCLUDING BUT NOT LIMITED TO iGALCON, GALCON LABS, GALCON FUSION, GALCON 2, AND ALL FUTURE GALCON RELATED SOFTWARE.  IN NO EVENT SHALL THE END USER DISPUTE OR DENY THIS FACT, AND esparano(TM) IS IN NO WAY RESPONSIBLE FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE DENIAL OF THE AFOREMENTIONED FACT DUE TO THE CRAB LORD'S WRATH AT HIS OR HER INSOLENCE.

P.S. Huge thanks to Milagre for developing the multiple-minigame framework.

]]

MODES = {
    {id="worldConquer",   name="World Conquer"},
    {id="berzerker", name="Berzerker"}
}

function pass()
end

function reset()
    OPTS = {
        sw = 640,
        sh = 480,    
    }
end
 
function init()
    --g2.speed = 30

    math.randomseed(os.time())

    COLORS = {0x555555,
        0x0000ff,0xff0000,
        0xffff00,0x00ffff,
        0xffffff,0xffbb00,
        0x99ff99,0xff9999,
        0xbb00ff,0xff88ff,
        0x9999ff,0x00ff00,
    }

    main_menu()
end

function loop(t)
    OPTS.time = OPTS.time + t
    over, victory = _ENV[OPTS.mode .. "_loop"](t)
    if over then
        g2.state = "pause"

        if victory then
            _ENV[OPTS.mode .. "_victory"](winner);
        else
            _ENV[OPTS.mode .. "_failure"](winner);
        end
    end
end

function fix(v,d,a,b)
    if (type(v) == "string") then v = tonumber(v) end
    if (type(v) ~= "number") then v = d end
    if v < a then v = a end
    if v > b then v = b end
    return v
end

function mknumber(v)
    v = tonumber(v)
    if v ~= nil then return v end
    return 0
end

function start()
    _ENV[OPTS.mode .. "_setup"]();
    _ENV[OPTS.mode .. "_init"]();
    OPTS.time = nil;
    get_ready();
    g2.state = "pause"
end

function event(e)
    if e["type"] == "onclick" and e["value"] then
        if string.find(e["value"],"mode:") ~= nil then
            OPTS.mode = string.sub(e["value"],6)
            _ENV[OPTS.mode .. "_menu"]()
        elseif string.find(e["value"],"newmap:") ~= nil then
            OPTS.data = string.sub(e["value"],8)
            start();
        elseif string.find(e["value"],"restart:") ~= nil then
            OPTS.data = string.sub(e["value"],9)
            start();
        elseif e["value"] == "newmap" or e["value"] == "restart" then
            start();
        elseif e["value"] == "home" then
            main_menu();
        elseif (e["value"] == "resume") then
            if OPTS.time == nil then OPTS.time = 0; end
            g2.state = "play"
        elseif (e["value"] == "quit") then
            g2.state = "quit"
        end

    elseif e["type"] == "pause" then
        paused();
        g2.state = "pause"
    end
end

function main_menu()
	g2.state = "menu"
    reset();

    local buttons = ""
    for _,v in ipairs(MODES) do
        buttons = buttons .. "<tr><td><input type='button' value='"..v["name"].."' onclick='mode:"..v["id"].."'/>"
    end

    g2.html = [[
        <table>
            <tr><td><h1>Galcon Frustration</h1>
            <tr><td><p>by esparano</p>
            <tr><td><p></p>
            <tr><td><p>Select a minigame:</p>
            ]]..buttons..[[
            <tr><td><p></p>
        </table>
    ]]
end

function get_ready()
    g2.html = 
    "<table>"..
    "<tr><td><h1>Get Ready!</h1>"..
    "<tr><td><input type='button' value='Tap to Begin' onclick='resume' />"..
    ""
end

function paused()
    local restart_opts = ""
    if OPTS.data then
        restart_opts = ":" .. OPTS.data
    end
    g2.html = 
    "<table>"..
    "<tr><td><input type='button' value='Resume'          onclick='resume' />"..
    "<tr><td><input type='button' value='Restart'         onclick='restart"..restart_opts.."' />"..
    "<tr><td><input type='button' value='New Challenge'   onclick='home' />"..
    "<tr><td><input type='button' value='Quit'            onclick='quit' />"..
    ""
end

function find_finish(user,from)
    local planets = g2.search("planet -team:"..user:team().." -neutral")
    local to = nil; local to_v = 0;
    for _i,p in ipairs(planets) do
        local d = from:distance(p)
        local v = -p.ships_value + p.ships_production - d * 0.20;
        if (to==nil or v > to_v) then
            to_v = v;
            to = p;
        end
    end
    return to
end
    
function find_normal(user,from)
    local planets = g2.search("planet -team:"..user:team());
    local to = nil; local to_v = 0;
    for _i,p in ipairs(planets) do
        local d = from:distance(p)
        local v = -p.ships_value + p.ships_production - d * 0.20;
        if (to==nil or v > to_v) then
            to_v = v;
            to = p;
        end
    end
    return to
end

function do_redirect(user,find)
    local fleets = g2.search("fleet owner:"..user);
    for _i,from in ipairs(fleets) do
        to = find(user,from)
        if (to ~= nil and to ~= from) then
            from:fleet_redirect(to)
        end
    end
end

function count_ships()
    local r = {}
    
    local items = g2.search("planet OR fleet")
    for _i,o in ipairs(items) do
        local uid = o.owner_n
        r[uid] = mknumber(r[uid]) + o.ships_value
    end
    
    return r
end

function is_winning(user)
    local r = count_ships()
    local totals = {0,0,0} --neutral,team,other
    local team = user:team()
    for n,t in ipairs(r) do
        local u = g2.item(n)
        if (u.user_neutral == 1) then
            totals[1] = totals[1] + t
        elseif (u:team() ~= team) then
            totals[3] = totals[3] + t
        else
            totals[2] = totals[2] + t
        end
    end
    return (totals[2] > totals[3]*2)
end

function find_from(user)
    local from = nil; local from_v = 0;
    local planets = g2.search("planet owner:"..user);
    for _i,p in ipairs(planets) do
        local v = p.ships_value
        if (v < 20) and OPTS.time > 3.5 then
            pass()
        elseif (v > from_v) then
            from_v = v;
            from = p;
        end
    end
    return from
end

---- {{{ BERZERKER

function berzerker_menu()
	g2.state = "menu"
    g2.html = [[
	    <table>
        <tr><td colspan=2><h2>Galcon Frustration:</h2>
	    <tr><td colspan=2><h2>Berzerker</h2>
	    <tr><td colspan=2><p>Mod by esparano</p>
	    <tr><td><p>&nbsp;</p>
	    <tr><td colspan=2><p>You start with 12000 ships.</p>
	    <tr><td colspan=2><p>They start with 0.</p>
	    <tr><td colspan=2><p>How can you possibly lose?</p>
	    <tr><td><p>&nbsp;</p>
	    <tr><td colspan=2>
	    <table><tr>
	    <td><input type='button'value="Bring it on." width=20 height=20 onclick='newmap' />
	    </table>
	    </table>
	]]
end

function berzerker_setup()
	OPTS.t    = 0
    OPTS.wait = 0.1
end

function berzerker_init()
    g2.game_reset();
    
    local neutral = g2.new_user("neutral", COLORS[1])
    neutral.user_neutral = 1
    neutral.ships_production_enabled = 0
    local a = 0
    for i=1,8 do
        a = a + 360/8*math.pi/180
        local d = 0
        for j=1,10 do
            d = d + 50
    	    planet = g2.new_planet(neutral, d*math.cos(a), d*math.sin(a), 100, 0);
   	    end
    end
    
    local player = g2.new_user("player", COLORS[2])
    OPTS.player = player
    OPTS.homes = {
        g2.new_planet(player, 0, 550, 100, 12000),
    }
    g2.player = player
    
    local enemy_team = g2.new_team("enemies",COLORS[3])
    a = 90*math.pi/180
    for i=4,10 do
        a = a + 360/8*math.pi/180
        local enemy = g2.new_user("bot", COLORS[i], enemy_team)
        planet = g2.new_planet(enemy, 550*math.cos(a), 550*math.sin(a), 100, 0);
    end
    
    g2.planets_settle()
end

function berzerker_loop(t)
	OPTS.t = OPTS.t + t
    if (OPTS.t >= OPTS.wait) then
        OPTS.t = OPTS.t - OPTS.wait
        local users = g2.search("user")
        for _i,user in ipairs(users) do
            if (user.title_value == "bot") then
                bot_berzerker(user)
            end
        end
    end
    
    local over = false
    local victory = false
    local winner = nil;
    local planets = g2.search("planet -neutral")
    for _i,p in ipairs(planets) do
        local team = p:owner():team()
        if (winner == nil) then winner = team end
        if (winner ~= team) then return over, victory end
    end
    over = true;
    if (winner ~= nil) then
        if (winner.has_player == 1) then
            victory = true
        else
           	victory = false
        end
    end
    return over, victory
end

function bot_berzerker(user)
	local perc = 100
    local find = find_normal
    if (is_winning(user)) then
        find = find_finish
    end
    local from = find_from(user)
    if (from ~= nil) then
        local to = find(user,from)
        if (to ~= nil) then
            from:fleet_send(perc,to)
        end
    end
    do_redirect(user,find)
end

function berzerker_victory() 
	worldConquer_victory()
end

function berzerker_failure() 
    worldConquer_failure()
end

--- }}} END BERZERKER

---- {{{ WORLD CONQUER

function worldConquer_menu()
	g2.state = "menu"
    g2.html = [[
	    <table>
        <tr><td colspan=2><h2>Galcon Frustration:</h2>
	    <tr><td colspan=2><h2>World Conquer</h2>
	    <tr><td colspan=2><p>Mod by esparano</p>
	    <tr><td><p>&nbsp;</p>
	    <tr><td colspan=2><p>Keep Calm and Fight On.</p>
	    <tr><td colspan=2><p>Last as long as possible.</p>
	    <tr><td><p>&nbsp;</p>
	    <tr><td colspan=2>
	    <table><tr>
	    <td><input type='button'value="Bring it on." width=20 height=20 onclick='newmap' />
	    </table>
	    </table>
	]]
end

function worldConquer_setup()
	OPTS.t    = 0
    OPTS.wait = 1.5
end

function worldConquer_init()
    g2.game_reset();
    
    local x1
    local y1
    local x2 = COASTLINE[1].x
    local y2 = COASTLINE[1].y
    for i=2,(# COASTLINE) do
    	x1 = x2
    	y1 = y2
    	x2 = COASTLINE[i].x
    	y2 = COASTLINE[i].y
    	if (COASTLINE[i].disconnect == nil) then
    		g2.new_line(0x44bbbb, 6*OPTS.sw*(y1 + 180)/360, 6*OPTS.sh*(-x1 + 90)/180, 6*OPTS.sw*(y2 + 180)/360, 6*OPTS.sh*(-x2 + 90)/180)
    	end
  	end
    
    local neutral = g2.new_user("neutral", 0x999999)
    neutral.user_neutral = 1
    neutral.ships_production_enabled = 0
    for i=1,(# COUNTRIES) do
    	planet = g2.new_planet(neutral, 6*OPTS.sw*(COUNTRIES[i].long + 180)/360, 6*OPTS.sh*(-COUNTRIES[i].lat + 90)/180, COUNTRIES[i].pop/110000 + 10, COUNTRIES[i].pop/300000 + 5);
   	    planet.title_value = COUNTRIES[i].name
    end
    
    local player = g2.new_user("player", COLORS[2])
    OPTS.player = player
    OPTS.homes = {
        g2.new_planet(player, 6*OPTS.sw*(-78 + 180)/380, 6*OPTS.sw*(-55 + 90)/180, 200, 200),
    }
    g2.player = player
    
    local enemy = g2.new_user("bot", COLORS[3])
    g2.new_planet(enemy, 6*OPTS.sw*(130 + 180)/380, 6*OPTS.sw*(-50 + 90)/180, 50, 100)
    g2.planets_settle()
    
end

function worldConquer_loop(t)
	OPTS.t = OPTS.t + t
    if (OPTS.t >= OPTS.wait) then
        OPTS.t = OPTS.t - OPTS.wait
        local users = g2.search("user")
        for _i,user in ipairs(users) do
            if (user.title_value == "bot") then
                bot_worldConquer(user)
            end
        end
    end
    
    local over = false
    local victory = false
    local winner = nil;
    local planets = g2.search("planet -neutral")
    for _i,p in ipairs(planets) do
        local user = p:owner()
        if (winner == nil) then winner = user end
        if (winner ~= user) then return over, victory end
    end
    over = true;
    if (winner ~= nil) then
        if (winner.has_player == 1) then
            victory = true
        else
           	victory = false
        end
    end
    return over, victory
end

function worldConquer_victory() 
	g2.html =
    "<table>"..
    "<tr><td><h1>You are epic.</h1>"..
    "<tr><td><p>It took you a mere " .. string.format("%d", OPTS.time) .. " seconds.</p>"..
    "<tr><td>"..
    "<tr><td>"..
    "<tr><td><input type='button' value='Replay' onclick='restart' />"..
    "<tr><td><input type='button' value='Try Another Minigame' onclick='home' />"..
    "<tr><td><input type='button' value='Quit' onclick='quit' />"..
    "<tr><td><input type='button' value='Receive Cookie' onclick='' />"..
    ""
end

function worldConquer_failure() 
    g2.html = 
    "<table>"..
    "<tr><td><h1>You suck.</h1>"..
    "<tr><td><p>You only lasted " .. string.format("%d", OPTS.time) .. " seconds.</p>"..
    "<tr><td>"..
    "<tr><td>"..
    "<tr><td><input type='button' value='Try Again' onclick='restart' />"..
    "<tr><td><input type='button' value='Try Another Minigame' onclick='home' />"..
    "<tr><td><input type='button' value='RageQuit' onclick='quit' />"..
    "<tr><td><input type='button' value='Commit Suicide' onclick='' />"..
    ""
end

function bot_worldConquer(user)
	local perc = 65
    local find = find_normal
    if (is_winning(user)) then
        find = find_finish
    end
    local from = find_from(user)
    if (from ~= nil) then
        local to = find(user,from)
        if (to ~= nil) then
            from:fleet_send(perc,to)
        end
    end
    do_redirect(user,find)
end

COUNTRIES = {
	{name="ABIDJAN,COTE_DIVOIRE", pop=1929079.0, long=-4.02, lat=5.32},
	{name="ADANA,TURKEY", pop=1066544.0, long=35.32, lat=37.0},
	{name="ADDIS_ABABA,ETHIOPIA", pop=2316400.0, long=38.7, lat=9.05},
	{name="ADELAIDE,AUSTRALIA", pop=1071100.0, long=138.6, lat=-34.93},
	{name="AHMEDABAD,INDIA", pop=2876710.0, long=72.67, lat=23.0},
	{name="ALEPPO,SYRIAN_ARAB_REPUBLIC", pop=1542000.0, long=37.17, lat=36.23},
	{name="ALEXANDRIA,EGYPT", pop=3380000.0, long=29.92, lat=31.22},
	{name="ALGER,ALGERIA", pop=1523000.0, long=3.0, lat=36.75},
	{name="ALMATY,KAZAKSTAN", pop=1176000.0, long=76.92, lat=43.32},
	{name="ANKARA,TURKEY", pop=2837937.0, long=32.83, lat=39.92},
	{name="ANSHAN,CHINA", pop=2478650.0, long=122.95, lat=41.12},
	{name="BAGHDAD,IRAQ", pop=3841268.0, long=44.43, lat=33.33},
	{name="BAKU,AZERBAIJAN", pop=1149000.0, long=49.88, lat=40.37},
	{name="BANDUNG,INDONESIA", pop=2368200.0, long=107.57, lat=-6.95},
	{name="BANGALORE,INDIA", pop=2660088.0, long=77.58, lat=12.97},
	{name="BARCELONA,SPAIN", pop=1596190.0, long=2.17, lat=41.42},
	{name="BEIJING,CHINA", pop=7362426.0, long=116.4, lat=39.93},
	{name="BELÉM,BRAZIL", pop=1297592.0, long=-48.48, lat=-1.45},
	{name="BELO_HORIZONTE,BRAZIL", pop=2060804.0, long=-43.9, lat=-19.9},
	{name="BEOGRAD,SERBIA", pop=1136786.0, long=20.5, lat=44.83},
	{name="BERLIN,GERMANY", pop=3472009.0, long=13.42, lat=52.53},
	{name="BHOPAL,INDIA", pop=1062771.0, long=77.47, lat=23.28},
	{name="BIRMINGHAM,UNITED_KINGDOM", pop=1008381.0, long=-1.92, lat=52.47},
	{name="BOMBAY,INDIA", pop=9925891.0, long=72.85, lat=18.98},
	{name="BOTOU,CHINA", pop=1228772.0, long=110.05, lat=40.6},
	{name="BRASILIA,BRAZIL", pop=1601094.0, long=-47.67, lat=-15.92},
	{name="BRISBANE,AUSTRALIA", pop=1421600.0, long=153.0, lat=-27.5},
	{name="BUCURESTI,ROMANIA", pop=2060551.0, long=26.17, lat=44.45},
	{name="BUDAPEST,HUNGARY", pop=1962855.0, long=19.08, lat=47.5},
	{name="BUENOS_AIRES,ARGENTINA", pop=2960976.0, long=-58.5, lat=-34.67},
	{name="BURSA,TURKEY", pop=1016760.0, long=29.07, lat=40.2},
	{name="BUSAN,SOUTH_KOREA", pop=3802319.0, long=129.02, lat=35.1},
	{name="CÓRDOBA,ARGENTINA", pop=1148305.0, long=-64.17, lat=-31.33},
	{name="CAIRO,EGYPT", pop=6800000.0, long=31.25, lat=30.05},
	{name="CALCUTTA,INDIA", pop=4399819.0, long=88.33, lat=22.5},
	{name="CARACAS,VENEZUELA", pop=1964846.0, long=-66.93, lat=10.54},
	{name="CARTAGENA,COLOMBIA", pop=1369331.0, long=-75.5, lat=10.4},
	{name="CHANGCHUN,CHINA", pop=2980870.0, long=125.35, lat=43.87},
	{name="CHANGDE,CHINA", pop=1409748.0, long=111.68, lat=29.03},
	{name="CHANGSHA,CHINA", pop=1328950.0, long=112.97, lat=28.2},
	{name="CHAOZHOU,CHINA", pop=1293507.0, long=116.63, lat=23.68},
	{name="CHELYABINSK,RUSSIA", pop=1087609.0, long=61.42, lat=55.2},
	{name="CHENGDU,CHINA", pop=3483834.0, long=104.07, lat=30.67},
	{name="CHICAGO,UNITED_STATES", pop=2731743.0, long=-87.75, lat=41.83},
	{name="CHINIOT,PAKISTAN", pop=2952689.0, long=73.0, lat=31.67},
	{name="CHONGQING,CHINA", pop=3122704.0, long=106.58, lat=29.57},
	{name="CURITIBA,BRAZIL", pop=1364320.0, long=-49.42, lat=-25.42},
	{name="DAEGU,SOUTH_KOREA", pop=2255805.0, long=128.6, lat=35.87},
	{name="DAEJEAN,SOUTH_KOREA", pop=1182932.0, long=127.43, lat=36.33},
	{name="DAKAR,SENEGAL", pop=1641358.0, long=-17.48, lat=14.72},
	{name="DALIAN,CHINA", pop=3473832.0, long=121.65, lat=38.92},
	{name="DALLAS,UNITED_STATES", pop=1022830.0, long=-96.83, lat=32.83},
	{name="DAMASCUS,SYRIAN_ARAB_REPUBLIC", pop=1549000.0, long=36.32, lat=33.5},
	{name="DAR_ES_SALAAM,TANZANIA", pop=1096000.0, long=39.3, lat=-6.88},
	{name="DATONG,CHINA", pop=1277310.0, long=113.3, lat=40.08},
	{name="DELHI,INDIA", pop=7206704.0, long=77.23, lat=28.67},
	{name="DENGZHOU,CHINA", pop=1391056.0, long=112.08, lat=32.68},
	{name="DEYANG,CHINA", pop=1286664.0, long=104.4, lat=31.13},
	{name="DNEPROPETROVSK,UKRAINE", pop=1185500.0, long=35.0, lat=48.48},
	{name="DONETSK,UKRAINE", pop=1121200.0, long=37.83, lat=48.0},
	{name="DONGSHAN,CHINA", pop=1413036.0, long=113.37, lat=22.52},
	{name="DONGWAN,CHINA", pop=1736869.0, long=113.73, lat=23.03},
	{name="DOUALA,CAMEROON", pop=1029731.0, long=9.72, lat=4.07},
	{name="ECATEPEC,MEXICO", pop=1218135.0, long=-99.07, lat=19.58},
	{name="EKATERINOBURG,RUSSIA", pop=1277963.0, long=60.6, lat=56.85},
	{name="EL_JADIDA,MOROCCO", pop=1220000.0, long=-8.5, lat=33.21},
	{name="ESFAHAN,IRAN", pop=1220595.0, long=51.68, lat=32.68},
	{name="FENGCHENG,CHINA", pop=1018911.0, long=115.77, lat=28.2},
	{name="FORTALEZA,BRAZIL", pop=1846955.0, long=-38.58, lat=-3.75},
	{name="FUKUOKA,JAPAN", pop=1275165.0, long=130.5, lat=33.5},
	{name="FUSHUN,CHINA", pop=1388011.0, long=123.88, lat=41.87},
	{name="FUZHOU,CHINA", pop=1395739.0, long=119.3, lat=26.08},
	{name="GIZA,EGYPT", pop=2144000.0, long=30.85, lat=30.05},
	{name="GUADALAJARA,MEXICO", pop=1626152.0, long=-103.33, lat=20.67},
	{name="GUANGZHOU,CHINA", pop=3918010.0, long=113.25, lat=23.12},
	{name="GUATEMALA,GUATEMALA", pop=1675589.0, long=-90.37, lat=14.63},
	{name="GUAYAQUIL,ECUADOR", pop=1508444.0, long=-79.92, lat=-2.22},
	{name="GUIKONG,CHINA", pop=1377391.0, long=111.72, lat=22.33},
	{name="GUIYANG,CHINA", pop=1664709.0, long=106.72, lat=26.58},
	{name="HAMBURG,GERMANY", pop=1705872.0, long=10.0, lat=53.55},
	{name="HANDAN,CHINA", pop=1769315.0, long=114.48, lat=36.58},
	{name="HANGZHOU,CHINA", pop=2589504.0, long=120.17, lat=30.25},
	{name="HANOI,VIETNAM", pop=1089760.0, long=105.87, lat=21.02},
	{name="HARARE,ZIMBABWE", pop=1189103.0, long=31.05, lat=-17.83},
	{name="HARBIN,CHINA", pop=3597404.0, long=126.65, lat=45.75},
	{name="HENGYANG,CHINA", pop=1814936.0, long=112.35, lat=26.97},
	{name="HIROSHIMA,JAPAN", pop=1106367.0, long=132.45, lat=34.38},
	{name="HO_CHI_MINH,VIETNAM", pop=3015743.0, long=106.72, lat=10.77},
	{name="HONG_KONG,HONG_KONG", pop=6190000.0, long=114.17, lat=22.27},
	{name="HOUSTON,UNITED_STATES", pop=1702086.0, long=-95.33, lat=29.83},
	{name="HUAIYIN,CHINA", pop=2613804.0, long=119.03, lat=33.58},
	{name="HYDERABAD,INDIA", pop=2964638.0, long=78.43, lat=17.37},
	{name="HYDERABAD,INDIA", pop=2964638.0, long=78.43, lat=17.37},
	{name="INCHON,SOUTH_KOREA", pop=2203102.0, long=126.67, lat=37.51},
	{name="ISTANBUL,TURKEY", pop=7774169.0, long=28.95, lat=41.03},
	{name="IZMIR,TURKEY", pop=2017699.0, long=27.17, lat=38.42},
	{name="JAKARTA,INDONESIA", pop=9160500.0, long=106.75, lat=-6.13},
	{name="JIAOZUO,CHINA", pop=1578461.0, long=113.22, lat=35.25},
	{name="JILIN,CHINA", pop=2251848.0, long=126.55, lat=43.85},
	{name="JINAN,CHINA", pop=2403946.0, long=117.0, lat=36.67},
	{name="JINHUA,CHINA", pop=2215666.0, long=119.65, lat=29.12},
	{name="JININ,CHINA", pop=1465656.0, long=116.55, lat=35.4},
	{name="JINXI,CHINA", pop=1350134.0, long=120.83, lat=40.75},
	{name="KABUL,AFGHANISTAN", pop=1424400.0, long=69.17, lat=34.5},
	{name="KANPUR,INDIA", pop=1874409.0, long=80.23, lat=26.45},
	{name="KAWASAKI,JAPAN", pop=1202069.0, long=139.68, lat=35.53},
	{name="KAZAN,RUSSIA", pop=1079800.0, long=49.17, lat=55.75},
	{name="KHARKOV,UKRAINE", pop=1615000.0, long=36.25, lat=50.0},
	{name="KHON_KAEN,THAILAND", pop=5876000.0, long=102.83, lat=16.42},
	{name="KIEV,UKRAINE", pop=2646100.0, long=30.5, lat=50.42},
	{name="KINSHASA,CONGO", pop=2664309.0, long=15.3, lat=-4.36},
	{name="KITAKYUSHU,JAPAN", pop=1019372.0, long=130.82, lat=33.87},
	{name="KOBE,JAPAN", pop=1518982.0, long=135.2, lat=34.68},
	{name="KUALA_LUMPUR,MALAYSIA", pop=1145075.0, long=101.7, lat=3.13},
	{name="KUNMING,CHINA", pop=1611969.0, long=102.68, lat=25.07},
	{name="KWANGCHU,SOUTH_KOREA", pop=1236312.0, long=126.92, lat=35.19},
	{name="KYOTO,JAPAN", pop=1448377.0, long=135.75, lat=35.03},
	{name="LA_HABANA,CUBA", pop=2175888.0, long=-82.37, lat=23.13},
	{name="LA_MATANZA,ARGENTINA", pop=1111811.0, long=-58.58, lat=-34.75},
	{name="LAGOS,NIGERIA", pop=1060848.0, long=3.47, lat=6.45},
	{name="LIMA,PERU", pop=6414500.0, long=-77.05, lat=-12.1},
	{name="LINYI,CHINA", pop=1590160.0, long=118.33, lat=35.08},
	{name="LIUPANSHUI,CHINA", pop=1844471.0, long=104.87, lat=26.55},
	{name="LONDON,UNITED_KINGDOM", pop=6962319.0, long=-0.17, lat=51.5},
	{name="LOS_ANGELES,UNITED_STATES", pop=3448613.0, long=-118.17, lat=34.0},
	{name="LUCKNOW,INDIA", pop=1619115.0, long=80.9, lat=26.83},
	{name="LUDHIANA,INDIA", pop=1042740.0, long=75.87, lat=30.93},
	{name="MÜNCHEN,GERMANY", pop=1244676.0, long=11.58, lat=48.13},
	{name="MADRAS,INDIA", pop=3841396.0, long=80.28, lat=13.08},
	{name="MADRID,SPAIN", pop=2976064.0, long=-3.71, lat=40.41},
	{name="MANAUS,BRAZIL", pop=1078277.0, long=-60.0, lat=-3.1},
	{name="MANILA,PHILIPPINES", pop=1728441.0, long=120.97, lat=14.62},
	{name="MARACAIBO,VENEZUELA", pop=1207513.0, long=-71.66, lat=10.73},
	{name="MASHHAD,IRAN", pop=1964489.0, long=59.57, lat=36.27},
	{name="MEDAN,INDONESIA", pop=1909700.0, long=98.65, lat=3.58},
	{name="MELBOURNE,AUSTRALIA", pop=3189200.0, long=144.97, lat=-37.75},
	{name="MEXICO_CITY,MEXICO", pop=8235744.0, long=-99.17, lat=19.42},
	{name="MIANYANG,CHINA", pop=1701059.0, long=104.77, lat=31.47},
	{name="MINSK,BELARUS", pop=1659700.0, long=27.5, lat=53.85},
	{name="MONTERREY,MEXICO", pop=1084696.0, long=-100.33, lat=25.67},
	{name="MONTEVIDEO,URUGUAY", pop=1360258.0, long=-56.17, lat=-34.87},
	{name="MONTREAL,CANADA", pop=1017666.0, long=-73.5833, lat=45.5},
	{name="MOSCOW,RUSSIA", pop=8436447.0, long=37.7, lat=55.75},
	{name="NAGOYA,JAPAN", pop=2153293.0, long=136.88, lat=35.13},
	{name="NAGPUR,INDIA", pop=1624752.0, long=79.2, lat=21.17},
	{name="NAIROBI,KENYA", pop=1162189.0, long=36.83, lat=-1.28},
	{name="NANJIANG,CHINA", pop=2610594.0, long=118.78, lat=32.05},
	{name="NANTONG,CHINA", pop=1602029.0, long=118.17, lat=26.63},
	{name="NETZAHUALCÓYOTL,MEXICO", pop=1255456.0, long=99.03, lat=19.42},
	{name="NEW_YORK,UNITED_STATES", pop=7333253.0, long=-73.98, lat=40.75},
	{name="NINGBO,CHINA", pop=3350851.0, long=121.55, lat=29.88},
	{name="NIZHNY_NOVGOROD,RUSSIA", pop=1382115.0, long=44.0, lat=56.33},
	{name="NOVOSIBIRSK,RUSSIA", pop=1370544.0, long=83.08, lat=55.07},
	{name="ODESSA,UKRAINE", pop=1086700.0, long=30.75, lat=46.5},
	{name="OKARA,PAKISTAN", pop=5180562.0, long=73.52, lat=30.82},
	{name="OMSK,RUSSIA", pop=1161991.0, long=73.37, lat=55.0},
	{name="OSAKA,JAPAN", pop=2575042.0, long=135.5, lat=34.67},
	{name="PÓRTO_ALEGRE,BRAZIL", pop=1280114.0, long=-51.17, lat=-30.05},
	{name="PADOVA,ITALY", pop=2693383.0, long=11.88, lat=45.4},
	{name="PALEMBANG,INDONESIA", pop=1352300.0, long=104.75, lat=-2.98},
	{name="PALERMO,ITALY", pop=1054601.0, long=13.38, lat=38.1},
	{name="PALMIRA,COLOMBIA", pop=1452392.0, long=-76.28, lat=3.55},
	{name="PARIS,FRANCE", pop=2152329.0, long=2.33, lat=48.87},
	{name="PARMA,ITALY", pop=1371008.0, long=10.32, lat=44.8},
	{name="PERM,RUSSIA", pop=1034201.0, long=56.17, lat=58.02},
	{name="PERTH,AUSTRALIA", pop=1221200.0, long=115.82, lat=-31.97},
	{name="PHILADELPHIA,UNITED_STATES", pop=1524249.0, long=-75.17, lat=40.0},
	{name="PHOENIX,UNITED_STATES", pop=1048949.0, long=-112.17, lat=33.5},
	{name="PINGDINGSHAN,CHINA", pop=1827593.0, long=113.3, lat=33.73},
	{name="PINXIANG,CHINA", pop=1388427.0, long=113.85, lat=27.62},
	{name="PRAHA,CZECH_REPUBLIC", pop=1216568.0, long=14.43, lat=50.1},
	{name="PUNE,INDIA", pop=1566651.0, long=73.97, lat=18.57},
	{name="QINGDO,CHINA", pop=5124868.0, long=120.32, lat=36.07},
	{name="QUEZON_CITY,PHILIPPINES", pop=1676644.0, long=121.03, lat=14.65},
	{name="QUITO,ECUADOR", pop=1100847.0, long=-78.5, lat=-0.23},
	{name="RAJSHAHI,BANGLADESH", pop=1363998.0, long=88.67, lat=24.4},
	{name="RECIFE,BRAZIL", pop=1314857.0, long=-34.88, lat=-8.06},
	{name="RIO_DE_JANEIRO,BRAZIL", pop=5547033.0, long=-43.28, lat=-22.88},
	{name="ROSTOV-NA-DONU,RUSSIA", pop=1013001.0, long=39.75, lat=47.25},
	{name="SAIDPUR,BANGLADESH", pop=3397187.0, long=89.0, lat=25.8},
	{name="SALVADOR,BRAZIL", pop=2174072.0, long=-38.48, lat=-12.97},
	{name="SAMARA,RUSSIA", pop=1190191.0, long=50.17, lat=53.17},
	{name="SAN_DIEGO,UNITED_STATES", pop=1151977.0, long=-117.17, lat=32.83},
	{name="SANTIAGO,CHILE", pop=4229970.0, long=-70.67, lat=-33.5},
	{name="SANTO_DOMINGO,DOMINICAN_REPUBLIC", pop=2134779.0, long=-69.95, lat=18.5},
	{name="SAO_PAOLO,BRAZIL", pop=9842059.0, long=-46.63, lat=-23.55},
	{name="SAPPORO,JAPAN", pop=1744806.0, long=141.35, lat=43.08},
	{name="SEMARANG,INDONESIA", pop=1366500.0, long=110.48, lat=-6.97},
	{name="SEOUL,SOUTH_KOREA", pop=1.0776201E7, long=127.0, lat=37.5},
	{name="SHANGHAI,CHINA", pop=8205598.0, long=121.47, lat=31.23},
	{name="SHEIKHU_PURA,PAKISTAN", pop=1104209.0, long=74.13, lat=31.7},
	{name="SHENYANG,CHINA", pop=4655280.0, long=123.45, lat=41.8},
	{name="SHIJIANZHUANG,CHINA", pop=1372109.0, long=114.48, lat=38.05},
	{name="SHIRAZ,IRAN", pop=1042801.0, long=52.57, lat=29.63},
	{name="SINGAPORE,SINGAPORE", pop=2987000.0, long=103.85, lat=1.28},
	{name="SIPING,CHINA", pop=1395094.0, long=124.33, lat=43.17},
	{name="SOFIA,BULGARIA", pop=1113674.0, long=23.3, lat=42.67},
	{name="ST.PETERSBURG,RUSSIA", pop=4273001.0, long=30.42, lat=59.92},
	{name="SURABAYA,INDONESIA", pop=2701300.0, long=112.75, lat=-7.23},
	{name="SURAT,INDIA", pop=1498817.0, long=72.9, lat=21.17},
	{name="SUZHOU,CHINA", pop=3273010.0, long=120.62, lat=31.3},
	{name="SYDNEY,AUSTRALIA", pop=3713500.0, long=151.17, lat=-33.92},
	{name="TABRIZ,IRAN", pop=1166203.0, long=46.3, lat=38.08},
	{name="TAIAN,CHINA", pop=3824748.0, long=117.12, lat=36.2},
	{name="TAIPEI,TAIWAN", pop=1769568.0, long=121.45, lat=25.02},
	{name="TAIYAN,CHINA", pop=2224580.0, long=112.55, lat=37.87},
	{name="TANGER,MOROCCO", pop=2943000.0, long=-5.83, lat=35.74},
	{name="TANGERANG,INDONESIA", pop=1198300.0, long=106.62, lat=-6.18},
	{name="TANGSHAN,CHINA", pop=1484515.0, long=114.73, lat=37.33},
	{name="TASHKENT,UZBEKISTAN", pop=2094000.0, long=69.22, lat=41.27},
	{name="TBILISI,GEORGIA", pop=1268000.0, long=44.8, lat=41.72},
	{name="TEHRAN,IRAN", pop=6750043.0, long=51.43, lat=35.67},
	{name="TIANJIN,CHINA", pop=5804023.0, long=117.2, lat=39.13},
	{name="TOKYO,JAPAN", pop=8129377.0, long=139.75, lat=35.67},
	{name="TONGHUA,CHINA", pop=1199382.0, long=125.75, lat=41.68},
	{name="TULUA,COLOMBIA", pop=3974813.0, long=-76.2, lat=4.08},
	{name="UFA,RUSSIA", pop=1090644.0, long=55.97, lat=54.75},
	{name="UJUNG_PANDANG,INDONESIA", pop=1091800.0, long=119.47, lat=-5.15},
	{name="URUMQI,CHINA", pop=1160775.0, long=87.58, lat=43.8},
	{name="VADODARA,INDIA", pop=1031346.0, long=73.23, lat=22.32},
	{name="VALENCIA,VENEZUELA", pop=1034033.0, long=-67.98, lat=10.23},
	{name="VIENNA,AUSTRIA", pop=1560471.0, long=16.37, lat=48.22},
	{name="WARSAW,POLAND", pop=1643203.0, long=21.0, lat=52.25},
	{name="WEIFANG,CHINA", pop=3037535.0, long=119.1, lat=36.72},
	{name="WUHAN,CHINA", pop=3832536.0, long=114.27, lat=30.58},
	{name="WUXI,CHINA", pop=3181985.0, long=120.3, lat=31.58},
	{name="XIAN,CHINA", pop=2872539.0, long=108.9, lat=34.27},
	{name="XIANGFAN,CHINA", pop=2985978.0, long=112.13, lat=32.02},
	{name="XUZHOU,CHINA", pop=1827306.0, long=117.18, lat=34.27},
	{name="YANGON,MYANMAR", pop=2513023.0, long=96.17, lat=16.78},
	{name="YANGZHOU,CHINA", pop=2769300.0, long=119.43, lat=32.4},
	{name="YANTAI,CHINA", pop=3204669.0, long=121.4, lat=37.53},
	{name="YOKOHAMA,JAPAN", pop=3300513.0, long=139.63, lat=35.47},
	{name="YULIN,CHINA", pop=1323410.0, long=110.15, lat=22.63},
	{name="ZAOZHUANG,CHINA", pop=3191974.0, long=117.57, lat=34.88},
	{name="ZHENGZHOU,CHINA", pop=1752374.0, long=113.67, lat=34.75},
	{name="ZHUZHOU,CHINA", pop=1521879.0, long=113.15, lat=27.83},
	{name="ZIBO,CHINA", pop=2484206.0, long=118.05, lat=36.8}	
}

COASTLINE = {
{x=40.77317942900004, y=24.211192254000082, name="START_EUROPE"},
{x=40.31049225500004, y=24.01921634200005},
{x=40.093817450000074, y=23.793304884000065},
{x=40.44944896000004, y=22.867442254000082},
{x=40.04987213700008, y=22.583181186000047},
{x=39.16811758000006, y=23.213552280000044},
{x=38.94961172100005, y=22.94971764400009},
{x=38.63971588700008, y=23.13331139400009},
{x=38.05927155200004, y=24.00489342500009},
{x=37.80906810100004, y=23.777354363000086},
{x=37.52456289300005, y=22.734629754000082},
{x=36.788275458000044, y=23.091075066000087},
{x=36.808579820000034, y=22.736664259000065},
{x=36.90639883000006, y=22.20020592500009},
{x=37.28392161700009, y=21.682302280000044},
{x=38.330511786000045, y=21.915537957000083},
{x=38.06240469000005, y=23.10092207100007},
{x=38.293036200000074, y=22.588633660000085},
{x=38.31395091400009, y=21.535411004000082},
{x=40.908270575000074, y=19.381195509000065},
{x=41.602728583000044, y=19.596039259000065},
{x=42.36627838700008, y=18.653168165000068},
{x=42.77094147300005, y=17.755137566000087},
{x=42.95888906500005, y=17.515635613000086},
{x=43.49079010600008, y=16.03956139400009},
{x=43.82123444200005, y=15.673024936000047},
{x=44.53607819200005, y=15.080088738000086},
{x=45.00482819200005, y=14.075694207000083},
{x=44.34780508000006, y=12.325450066000087},
{x=42.07884349200003, y=14.75749759200005},
{x=40.67047760600008, y=17.955821160000085},
{x=40.240301825000074, y=17.91114342500009},
{x=39.55267975500004, y=16.84620201900009},
{x=37.924709377000056, y=16.015879754000082},
{x=39.86859772300005, y=15.781911655000044},
{x=40.60541413000004, y=14.327972852000073},
{x=41.315619208000044, y=12.992686394000089},
{x=42.40273672100005, y=11.094493035000085},
{x=44.05548737200007, y=9.98414147200009},
{x=44.360541083000044, y=8.576670769000089},
{x=43.66010941900004, y=7.177651002000061},
{x=43.16056742200004, y=6.526714449000053},
{x=43.207739925000055, y=5.514335724000091},
{x=43.57908763200004, y=4.694590691000087},
{x=42.45648834800005, y=3.175873243000069},
{x=41.23810455900008, y=1.863780144000089},
{x=39.63104889500005, y=-0.243723110999952},
{x=38.36343008000006, y=-0.445871548999946},
{x=37.498114325000074, y=-1.430368618999921},
{x=36.685980536000045, y=-2.684356248999961},
{x=36.247381903000075, y=-5.296620245999918},
{x=36.733303127000056, y=-6.437896287999934},
{x=37.014105536000045, y=-7.94953365799995},
{x=37.87809298600007, y=-8.79895597399991},
{x=38.65888092700004, y=-9.11392167899993},
{x=39.40582171900007, y=-9.279244908999942},
{x=40.81118398600006, y=-8.670521613999938},
{x=42.16083405200004, y=-8.846994594999956},
{x=42.46185944200005, y=-8.826527472999942},
{x=42.82330963700008, y=-9.106190558999913},
{x=43.23875560100004, y=-9.004058397999927},
{x=43.48773834800005, y=-8.247547980999911},
{x=43.76666901200008, y=-7.672596808999913},
{x=43.584051825000074, y=-6.249989386999914},
{x=43.44094472900008, y=-4.023426886999914},
{x=43.43077220300006, y=-3.501698370999918},
{x=43.30959707200009, y=-2.065541144999941},
{x=45.54633209800005, y=-1.057362433999913},
{x=47.29328034100007, y=-2.006581183999913},
{x=47.47870514500005, y=-3.098133917999917},
{x=47.734035549000055, y=-3.375477667999917},
{x=48.304673570000034, y=-4.54320227799991},
{x=48.66933828300006, y=-3.94953365799995},
{x=48.81635163000004, y=-3.208729620999918},
{x=49.13458893400008, y=-1.589019334999932},
{x=49.68016185100004, y=-1.707753058999913},
{x=49.47866445500006, y=0.476573113000086},
{x=50.863714911000045, y=1.580821160000085},
{x=51.38422272300005, y=4.249034050000091},
{x=51.61680735900006, y=4.186127149000072},
{x=52.93569570500006, y=4.97828209700009},
{x=53.46735260600008, y=7.027679884000065},
{x=53.53620026200008, y=8.55209394600007},
{x=56.558294989000046, y=9.302256707000083},
{x=57.04930247600004, y=9.580821160000085},
{x=56.60488515800006, y=8.555430535000085},
{x=57.003241278000075, y=10.338715040000068},
{x=55.80341217700004, y=10.042165561000047},
{x=55.27163320500006, y=9.694183790000068},
{x=54.90729401200008, y=9.617523634000065},
{x=54.42617422100005, y=10.191742384000065},
{x=54.177435614000046, y=12.110606316000087},
{x=53.75527578300006, y=14.037445509000065},
{x=53.899318752000056, y=14.438161655000044},
{x=54.367254950000074, y=16.29623457100007},
{x=54.43060944200005, y=18.603770379000082},
{x=55.473089911000045, y=21.10840905000009, name="END_EUROPE"},

{x=-19.684747002999927, y=44.41293379000007, name="START_MADAGASCAR", disconnect=true},
{x=-16.65553150799991, y=44.439789259000065},
{x=-15.34937916499996, y=47.06495201900009},
{x=-14.82236093499995, y=47.51449629000007},
{x=-13.516778252999927, y=48.51286868600005},
{x=-12.45476653399993, y=48.77084394600007},
{x=-13.33131275799991, y=49.972666863000086},
{x=-15.408623955999929, y=50.47999108200008},
{x=-24.814060153999947, y=47.14283287900008},
{x=-25.210544528999947, y=44.40943444100009},
{x=-21.199395440999922, y=43.82276451900009},
{x=-19.684747002999927, y=44.41293379000007, name="END_MADAGASCAR"},

{x=17.37490469000005, y=122.24634850400003, name="START_ISLANDS2", disconnect=true},
{x=16.257473049000055, y=122.20337975400003},
{x=15.437689520000049, y=121.47291100400003},
{x=13.01235586100006, y=123.31006920700008},
{x=13.73578522300005, y=122.51246178500003},
{x=13.949042059000078, y=121.77800540500004},
{x=13.816107489000046, y=120.90870201900009},
{x=14.74909088700008, y=120.84392337300005},
{x=16.27952708500004, y=120.34864342500009},
{x=18.219631252000056, y=120.52084394600001},
{x=18.32021719000005, y=122.03956139400009},
{x=17.37490469000005, y=122.24634850400003},

{x=8.43195221600007, y=126.38379967500009, disconnect=true},
{x=9.408351955000057, y=125.46631920700008},
{x=8.61587148600006, y=124.77149498800009},
{x=7.963364976000037, y=123.67644290500004},
{x=8.524237372000073, y=123.30095462300005},
{x=7.513657945000091, y=122.06999759200005},
{x=7.410345770000049, y=124.18775475400003},
{x=6.58234284100007, y=125.42188561300009},
{x=6.789740302000041, y=126.22575931100005},
{x=8.43195221600007, y=126.38379967500009, name="END_ISLANDS2"},

{x=7.66860586100006, y=81.70069420700008, disconnect=true},
{x=8.093166408000059, y=81.43563886800007},
{x=8.545803127000056, y=81.24976647200003},
{x=9.14516836100006, y=80.88217207100007},
{x=9.779038804000038, y=79.93132571700005},
{x=9.596380927000041, y=80.05298912900008},
{x=8.214504299000055, y=79.81291751400005},
{x=7.602728583000044, y=79.77857506600003},
{x=6.00657786700009, y=80.26002037900008},
{x=6.725043036000045, y=81.81397545700008},

{x=32.720363674000055, y=131.86215254, name="START_JAPAN", disconnect=true},
{x=33.20034414300005, y=131.87224368600005},
{x=32.78489817900004, y=129.78467858200008},
{x=31.669989325000074, y=130.2973738940001},
{x=31.412176825000074, y=131.24577884200005},
{x=33.95189325700005, y=134.69596250000006},
{x=34.34137604400007, y=133.1821395190001},
{x=34.362494208000044, y=132.49683678500003},
{x=34.00263092700004, y=131.59742272200003},
{x=33.99941640800006, y=130.913422071},
{x=34.40839264500005, y=131.222015821},
{x=35.355128136000076, y=132.67144646200006},
{x=35.593817450000074, y=134.34262129},
{x=35.50169505400004, y=135.37378991000003},
{x=35.625981151000076, y=135.9710528820001},
{x=37.13740149600005, y=136.67644378300008},
{x=37.09540888600009, y=137.02429826600007},
{x=40.95713361300005, y=140.861488388},
{x=41.19595586000008, y=140.76555366700006},
{x=38.446600653000075, y=141.50798587300005},
{x=37.75018952000005, y=141.008148634},
{x=35.133402885000066, y=140.303349485},
{x=34.691473700000074, y=138.7465926440001},
{x=34.59077881500008, y=137.02359643},
{x=33.781998007000084, y=135.23258264800006},
{x=34.67490879400009, y=135.186030437},
{x=34.57786692900004, y=133.925140821},
{x=32.720363674000055, y=131.86215254},

{x=43.244574286000045, y=145.49195397200003, disconnect=true},
{x=44.079413153000075, y=144.2558699880001},
{x=45.40208567900004, y=141.704437696},
{x=43.20062897300005, y=140.78630618600005},
{x=42.01113515800006, y=140.1118270190001},
{x=41.88263580900008, y=141.08326256600003},
{x=42.10537344000005, y=142.95972741000003},
{x=43.08820221600007, y=145.12273196700005},
{x=43.244574286000045, y=145.49195397200003},

{x=53.69757721600007, y=142.80005944100003, disconnect=true},
{x=53.09829336100006, y=141.88982181100005},
{x=51.61245351800005, y=141.86670983200008},
{x=49.623032945000034, y=142.14771569100003},
{x=46.62636953300006, y=141.84009850400003},
{x=48.807806708000044, y=142.92009524800005},
{x=51.50836823100008, y=143.3379012380001},
{x=52.556708075000074, y=143.2353621750001},
{x=53.46674225500004, y=143.1142684250001},
{x=54.40298086100006, y=142.75180097700002},
{x=53.69757721600007, y=142.80005944100003, name="END_JAPAN"},

{x=5.403021552000041, y=118.92172285200002, name="START_INDONESIA", disconnect=true},
{x=5.69407786700009, y=117.95199629000001},
{x=6.371568101000094, y=117.73308353000004},
{x=6.941310940000051, y=117.07667076900009},
{x=6.368841864000046, y=116.33838951900009},
{x=5.225490627000056, y=115.59864342500009},
{x=4.601141669000071, y=113.99878991000003},
{x=2.344794012000079, y=111.35840905000003},
{x=1.658107815000051, y=110.49284915500004},
{x=1.86790599200009, y=109.33643639400003},
{x=0.575018622000073, y=108.93116295700003},
{x=-0.587334893999923, y=109.37452233200003},
{x=-0.967543226999908, y=109.74048912900003},
{x=-2.701918226999908, y=110.22315514400003},
{x=-2.868096612999921, y=111.69874108200003},
{x=-3.176853122999944, y=113.24122155000009},
{x=-4.132256768999923, y=114.60661868600005},
{x=-3.005466403999947, y=116.26335696700005},
{x=-1.77898528399993, y=116.37256920700008},
{x=-1.121840101999908, y=117.06519616000003},
{x=-0.50904713299991, y=117.47494550900001},
{x=0.596380927000041, y=117.68441816500001},
{x=0.846380927000041, y=118.66277103000004},
{x=1.783107815000051, y=118.04615319100003},
{x=2.901597398000092, y=117.73633873800009},
{x=3.576971747000073, y=117.20533287900003},
{x=4.042954820000091, y=117.48894290500004},
{x=4.361517645000049, y=118.50342858200008},
{x=4.96743398600006, y=118.56023196700005},
{x=5.403021552000041, y=118.92172285200002},

{x=0.738104559000078, y=102.55193118600005, disconnect=true},
{x=0.345648505000042, y=103.63428795700003},
{x=-0.424574476999908, y=103.59050540500004},
{x=-1.040785414999959, y=104.38184655000003},
{x=-2.573988539999959, y=104.70923912900008},
{x=-2.552341403999947, y=105.60840905000003},
{x=-4.848565362999921, y=105.87240644600007},
{x=-5.609470309999949, y=104.79078209700003},
{x=-4.971774997999944, y=103.73292076900003},
{x=-2.61687590899993, y=101.13982181100005},
{x=-0.977146091999941, y=100.35303795700003},
{x=0.818752346000053, y=99.00611412900008},
{x=2.109930731000077, y=98.20948326900003},
{x=4.225653387000079, y=96.01002037900008},
{x=5.615301825000074, y=95.34522545700003},
{x=4.532782294000071, y=98.08961022200003},
{x=3.204738674000055, y=99.68531334700003},
{x=2.572088934000078, y=100.29957116000003},
{x=0.738104559000078, y=102.55193118600005},

{x=-8.127618096999925, y=110.54908287900003, disconnect=true},
{x=-8.256524346999925, y=111.79200280000003},
{x=-8.44703541499996, y=113.59693444100003},
{x=-8.491794528999947, y=114.40512129000001},
{x=-7.663018487999921, y=113.05600019600001},
{x=-6.881117445999905, y=110.19353274800005},
{x=-6.206397675999938, y=107.86396034600006},
{x=-5.917738539999959, y=105.99724368600005},
{x=-7.570433912999931, y=107.59096020900006},
{x=-8.127618096999925, y=110.54908287900003},

{x=0.001532294000071, y=119.71583092500009, disconnect=true},
{x=0.898586330000057, y=120.61784915500004},
{x=1.02407461100006, y=122.01221764400009},
{x=0.927883205000057, y=123.38331139400009},
{x=0.35187409100007, y=124.06853274800005},
{x=-1.805759372999944, y=121.37134850400003},
{x=-3.12859465899993, y=122.43384850400003},
{x=-3.746270440999922, y=122.45337975400003},
{x=-5.532403252999927, y=120.27344811300009},
{x=-4.698011976999908, y=119.52515709700003},
{x=-3.546807549999926, y=118.91684004000001},
{x=-1.961521091999941, y=119.30404707100001},
{x=0.001532294000071, y=119.71583092500009},

{x=-6.82000090899993, y=138.85670006600003, disconnect=true},
{x=-5.489434502999927, y=138.06421959700003},
{x=-4.991143487999921, y=137.16944420700008},
{x=-4.30592213299991, y=134.94312584700003},
{x=-3.913506768999923, y=134.4365340500001},
{x=-3.69068775799991, y=133.79908287900003},
{x=-3.386651299999926, y=133.52849368600005},
{x=-3.586358330999929, y=132.89625084700003},
{x=-2.914808851999908, y=131.996918165},
{x=-2.190850518999923, y=133.9170028000001},
{x=-2.295993747999944, y=132.67562910200002},
{x=-1.573337497999944, y=131.94906660200002},
{x=-1.433689059999949, y=130.95093834700003},
{x=-0.454034112999921, y=132.93677819100003},
{x=-1.202894789999959, y=134.19288170700008},
{x=-2.724867445999962, y=134.498383009},
{x=-3.281670830999929, y=135.56910241000003},
{x=-2.246270440999922, y=136.62647545700008},
{x=-2.155694268999923, y=139.38404381600003},
{x=-2.676446221999925, y=141.296234571},
{x=-3.68914153399993, y=143.8676863940001},
{x=-4.37468840899993, y=145.30689537900003},
{x=-5.537530205999929, y=146.16822350400003},
{x=-5.969414971999925, y=147.46656334700003},
{x=-6.954685153999947, y=146.95622806100005},
{x=-7.71843840899993, y=147.5859481130001},
{x=-8.730645440999922, y=148.48113040500004},
{x=-9.224216403999947, y=149.24488366000003},
{x=-9.43328215899993, y=147.07935631600003},
{x=-8.81812916499996, y=146.56592858200008},
{x=-7.937269789999959, y=145.58326256600003},
{x=-7.625258070999905, y=144.5644637380001},
{x=-7.730726820999905, y=144.2524520190001},
{x=-7.997979424999926, y=143.56316165500004},
{x=-8.232354424999926, y=142.36703535200002},
{x=-9.188653252999927, y=141.99398847700002},
{x=-8.124932549999926, y=139.63054446700005},
{x=-7.515232028999947, y=139.03589928500003},
{x=-6.972914320999905, y=139.14039147200003},
{x=-6.82000090899993, y=138.85670006600003},

{x=-42.25505950299993, y=173.87134850400003, disconnect=true},
{x=-41.32244231599992, y=174.24529056100005},
{x=-41.03142669099992, y=174.26441491000003},
{x=-41.1950822899999, y=174.02409915500004},
{x=-40.98593515399995, y=173.9223738940001},
{x=-41.24700286299992, y=173.30152428500003},
{x=-40.67782968499995, y=172.6796981130001},
{x=-41.3005510399999, y=172.10670006600003},
{x=-42.69296640399995, y=171.01246178500003},
{x=-43.23642343499995, y=170.4282332690001},
{x=-43.71697356599992, y=169.52214603000004},
{x=-44.27206796699994, y=168.1279403000001},
{x=-44.87859465899993, y=167.46998131600003},
{x=-45.13290780999995, y=167.012054884},
{x=-45.379571221999925, y=166.91000410200002},
{x=-45.985039971999925, y=166.5341903000001},
{x=-46.26913827899995, y=167.169281446},
{x=-46.53679778399993, y=168.32642662900003},
{x=-46.61801523199995, y=169.192718946},
{x=-45.801039320999905, y=170.74203535200002},
{x=-44.84628671699994, y=171.2114363940001},
{x=-44.058526299999926, y=171.86833743600005},
{x=-43.816664320999905, y=172.97868899800005},
{x=-43.42343515399995, y=172.6582137380001},
{x=-40.68173593499995, y=176.378265821},
{x=-41.22307708099993, y=174.675140821},
{x=-39.85947030999995, y=174.73503665500004},
{x=-38.06015390399995, y=174.83578535200002},
{x=-37.080010674999926, y=174.66521243600005},
{x=-36.24293385199991, y=174.487640821},
{x=-36.26816171699994, y=174.25066165500004},
{x=-35.424899997999944, y=173.4707137380001},
{x=-34.55632903399993, y=172.97046959700003},
{x=-34.998711846999925, y=173.59782962300005},
{x=-35.227227471999925, y=174.03052819100003},
{x=-36.81178150799991, y=174.73560631600003},
{x=-37.05901458099993, y=175.33269290500004},
{x=-36.529554945999905, y=175.35710696700005},
{x=-36.82577890399995, y=175.82781009200005},
{x=-37.896416924999926, y=178.41480553500003},
{x=-39.09734465899993, y=177.8662215500001},
{x=-42.25505950299993, y=173.87134850400003},

{x=65.99774811400005, y=-14.616932745999918, disconnect=true},
{x=65.53416575700004, y=-13.814442511999914},
{x=64.38890208500004, y=-14.720692511999914},
{x=64.37775299700007, y=-15.387847459999932},
{x=63.85028717700004, y=-16.757639126999948},
{x=63.77326080900008, y=-17.557118292999917},
{x=63.54645416900007, y=-18.29637610599991},
{x=63.68821849200003, y=-20.458241339999915},
{x=63.84845612200007, y=-21.076730923999946},
{x=63.96198151200008, y=-22.712880011999914},
{x=64.71580638200004, y=-22.36082923099991},
{x=65.69969310100004, y=-23.51162675699993},
{x=66.07660553600005, y=-23.577504035999937},
{x=65.17218659100007, y=-21.106190558999913},
{x=65.52850983300004, y=-20.55068925699993},
{x=66.16771067900004, y=-18.099720831999946},
{x=66.12433502800008, y=-16.933338995999918},
{x=66.41917552300004, y=-16.529286261999914},
{x=66.38686758000006, y=-15.697417772999927},
{x=65.99774811400005, y=-14.616932745999918},

{x=65.41699067300004, y=-179.4640020049999, disconnect=true},
{x=66.36099876000009, y=-178.5404711959999},
{x=65.45601814100007, y=-176.08759081399992},
{x=64.54910946800004, y=-174.07244961299992},
{x=64.41816172800009, y=-172.7145651069999},
{x=65.76121653900003, y=-170.57115637899992},
{x=66.31653175600007, y=-170.3851928839999},
{x=67.02138688200006, y=-174.49605294999992},
{x=66.33834554400005, y=-173.95256778599992},
{x=66.89940429200004, y=-174.76285577299993},
{x=67.53419921100004, y=-175.2183313719999},
{x=67.90177004200007, y=-176.20879499199992},
{x=68.29059479400007, y=-177.96479244699992},
{x=68.90863678600005, y=-179.57990475199992},
{x=65.41699067300004, y=-179.4640020049999, name="END_INDONESIA"},

{x=53.289129950000074, y=-6.109120245999918, name="START_GREAT_BRITAIN", disconnect=true},
{x=52.27260976800005, y=-6.36978105399993},
{x=52.251654364000046, y=-6.989898240999935},
{x=51.911810614000046, y=-7.922230597999942},
{x=51.70661041900007, y=-8.469227667999917},
{x=51.50800202000005, y=-9.203968878999945},
{x=51.491522528000075, y=-9.812001105999911},
{x=51.63666413000004, y=-10.061024542999917},
{x=51.794175523000035, y=-10.325754360999952},
{x=52.12323639500005, y=-10.288563605999911},
{x=52.28473541900007, y=-9.849232550999943},
{x=52.713771877000056, y=-8.949370897999927},
{x=52.82135651200008, y=-9.459462042999917},
{x=53.22825755400004, y=-9.470082160999937},
{x=53.432521877000056, y=-9.860340949999909},
{x=53.60984935100004, y=-9.714100714999915},
{x=54.01166413000004, y=-9.884388800999943},
{x=54.24433014500005, y=-10.08702551999994},
{x=54.280585028000075, y=-9.219634568999936},
{x=54.33466217700004, y=-8.540516730999911},
{x=54.627386786000045, y=-8.485218878999945},
{x=54.901556708000044, y=-8.432606574999909},
{x=55.13898346600007, y=-7.856434699999909},
{x=55.07111237200007, y=-7.586415167999917},
{x=55.21857330900008, y=-6.942779100999928},
{x=54.875067450000074, y=-5.82095292899993},
{x=54.42560455900008, y=-5.571685350999928},
{x=54.09975820500006, y=-6.24250240799995},
{x=53.289129950000074, y=-6.109120245999918},

{x=56.36286041900007, y=-2.809193488999938, disconnect=true},
{x=56.02228424700007, y=-2.863636847999942},
{x=55.29364655200004, y=-1.565744594999956},
{x=54.47736237200007, y=-0.562977667999917},
{x=53.683579820000034, y=-0.544748501999948},
{x=52.96816640800006, y=0.875498894000089},
{x=51.880560614000046, y=1.282237175000091},
{x=51.47703685100004, y=0.441661004000082},
{x=50.80963776200008, y=0.354258660000085},
{x=50.91815827000005, y=-1.466420050999943},
{x=50.55756256700005, y=-2.428700324999909},
{x=50.21588776200008, y=-3.740386522999927},
{x=50.33588288000004, y=-4.711333787999934},
{x=50.09975820500006, y=-5.540638800999943},
{x=50.62799713700008, y=-4.765044725999928},
{x=51.34666575700004, y=-2.987782355999911},
{x=51.476223049000055, y=-3.663197394999941},
{x=51.677801825000074, y=-4.075428839999915},
{x=51.706366278000075, y=-4.892648891999954},
{x=52.007473049000055, y=-4.942534959999932},
{x=52.882473049000055, y=-4.131988084999932},
{x=53.22907135600008, y=-4.066883917999917},
{x=53.39276764500005, y=-2.980824347999942},
{x=54.08869049700007, y=-3.147450324999909},
{x=54.94717031500005, y=-3.039540167999917},
{x=54.86627838700008, y=-4.200550910999937},
{x=54.728501695000034, y=-4.954701300999943},
{x=55.619208075000074, y=-4.721994594999956},
{x=55.98867422100005, y=-4.921457485999952},
{x=56.12921784100007, y=-5.228138800999943},
{x=55.38027578300006, y=-5.526682094999956},
{x=55.977728583000044, y=-5.674875454999949},
{x=56.54877350500004, y=-5.085682745999918},
{x=56.74445221600007, y=-5.23892167899993},
{x=56.72907135600008, y=-6.229562954999949},
{x=57.044175523000035, y=-5.710926886999914},
{x=57.37099844000005, y=-5.501291469999956},
{x=57.65509674700007, y=-5.81118730399993},
{x=57.91819896000004, y=-5.39289303299995},
{x=58.186712958000044, y=-5.329335089999915},
{x=58.43650950700004, y=-4.995594855999911},
{x=58.54637278900003, y=-4.306996222999942},
{x=58.451157945000034, y=-3.066761847999942},
{x=57.86542389500005, y=-3.786040818999936},
{x=57.54995351800005, y=-4.067290818999936},
{x=57.68109772300005, y=-1.982167120999918},
{x=56.45172760600008, y=-3.060617641999954, name="END_GREAT_BRITAIN"},

{x=82.64166901200008, y=-41.82909094999991, name= "START_GREENLAND", disconnect=true},
{x=82.15241120000007, y=-44.65078691299993},
{x=82.31617530600005, y=-48.36040876633327},
{x=82.02122630400004, y=-53.01427161399994},
{x=82.37270742400005, y=-54.440744594999956},
{x=81.68008047100005, y=-58.12474524599992},
{x=81.50885651200008, y=-60.78685462099992},
{x=80.99624258000006, y=-64.75352942599994},
{x=80.23509349200003, y=-67.43687903599994},
{x=79.82977936400005, y=-64.99819902299993},
{x=79.13833242400005, y=-66.47109941299993},
{x=78.83966705900008, y=-68.91136633999992},
{x=78.22748444200005, y=-72.78555253799993},
{x=77.70209381700005, y=-69.49559485599991},
{x=77.37685781500005, y=-66.60081946499992},
{x=76.90387604400007, y=-70.82957923099991},
{x=76.39032623900005, y=-69.62751217399995},
{x=76.21491120000007, y=-66.76032467399995},
{x=76.13515859600005, y=-65.05215410099993},
{x=76.22540924700007, y=-63.64826412699995},
{x=75.99917226800005, y=-60.733631964999915},
{x=75.71149323100008, y=-59.07681230399993},
{x=75.05182526200008, y=-58.14159094999991},
{x=74.59454987200007, y=-56.590728318999936},
{x=74.22699616100005, y=-56.164173956999946},
{x=73.75413646000004, y=-55.891468878999945},
{x=73.17059967700004, y=-55.350453253999945},
{x=72.73777903900003, y=-54.75377356699994},
{x=72.46092357000003, y=-55.304351365999935},
{x=72.11395905200004, y=-54.58185787699995},
{x=71.35366445500006, y=-54.78107662699995},
{x=72.25189850500004, y=-53.84117591099994},
{x=71.79694245000007, y=-53.00182044199994},
{x=71.19041575700004, y=-52.553578253999945},
{x=71.02142975500004, y=-51.91486568899995},
{x=70.59951406500005, y=-50.91234290299991},
{x=70.32440827000005, y=-50.61156165299991},
{x=70.01105377800008, y=-51.92491614499994},
{x=69.79140859600005, y=-50.27586829299992},
{x=69.49225495000007, y=-50.68517005099994},
{x=69.22500234600005, y=-50.450510219999956},
{x=68.92438385600008, y=-50.320423956999946},
{x=68.83417389500005, y=-51.20645097599993},
{x=68.56427643400008, y=-51.34064693899995},
{x=68.37152741100005, y=-52.910227016999954},
{x=68.22846100500004, y=-52.444203253999945},
{x=68.21482982000003, y=-51.11424719999991},
{x=67.97833893400008, y=-50.86082923099991},
{x=68.15053945500006, y=-53.17931067599994},
{x=67.89118073100008, y=-53.15021725199995},
{x=67.65062083500004, y=-53.589182094999956},
{x=67.78978099200003, y=-51.73155676999994},
{x=67.89618561400005, y=-51.05492102799991},
{x=67.74188873900005, y=-50.970041469999956},
{x=67.51593659100007, y=-50.77513587099992},
{x=67.49750397300005, y=-53.575835740999935},
{x=67.13939036700003, y=-51.11314856699994},
{x=67.13922760600008, y=-53.88975989499994},
{x=66.90395742400005, y=-52.302072719999956},
{x=66.67291901200008, y=-53.05406653599994},
{x=66.44334544500003, y=-53.58503170499995},
{x=67.01642487200007, y=-50.57127844999991},
{x=66.38930898600006, y=-52.48338782499991},
{x=65.97418854400007, y=-52.23810787699995},
{x=65.80878327000005, y=-52.877837693999936},
{x=65.47386302300004, y=-52.60118567599994},
{x=65.67560455900008, y=-51.76414954299992},
{x=65.28461334800005, y=-52.19859778599994},
{x=64.92373281500005, y=-52.05728105399993},
{x=65.01044342700004, y=-51.40477454299992},
{x=64.51848161100008, y=-52.12125145999994},
{x=64.65159739800004, y=-51.36583411399994},
{x=65.08283112200007, y=-50.94273841099994},
{x=64.82843659100007, y=-50.195423956999946},
{x=64.66571686400005, y=-50.44994055899991},
{x=64.42401764500005, y=-50.68858801999994},
{x=64.09658437700006, y=-51.47032630099994},
{x=64.07477448100008, y=-51.032460089999915},
{x=63.541001695000034, y=-51.42251542899993},
{x=63.45929596600007, y=-51.06094316299993},
{x=63.16693756700005, y=-50.93228105399993},
{x=62.99315013200004, y=-50.27806555899991},
{x=62.80801015800006, y=-50.14826412699995},
{x=62.266791083000044, y=-49.87328040299991},
{x=62.09096914300005, y=-49.34683183499993},
{x=62.06976959800005, y=-48.85383053299995},
{x=61.958319403000075, y=-48.83922278599994},
{x=61.56102122600004, y=-49.29857337099992},
{x=61.389105536000045, y=-48.572336391999954},
{x=61.262884833000044, y=-48.18211829299992},
{x=61.010646877000056, y=-48.16673743399991},
{x=60.86090729400007, y=-47.99372311099995},
{x=60.87909577000005, y=-47.044178839999915},
{x=61.056301174000055, y=-46.44432532499991},
{x=61.33490631700005, y=-45.767241990999935},
{x=61.246527411000045, y=-45.50503495999993},
{x=60.721584377000056, y=-46.060658331999946},
{x=60.78359609600005, y=-45.62800045499995},
{x=60.56232330900008, y=-45.622303839999915},
{x=60.58478424700007, y=-45.21776282499991},
{x=60.251206773000035, y=-45.14305579299992},
{x=60.087551174000055, y=-45.17284094999991},
{x=60.26870351800005, y=-44.22411048099991},
{x=60.13532135600008, y=-43.21328691299993},
{x=60.380845445000034, y=-43.47765051999994},
{x=60.572333075000074, y=-42.889881964999915},
{x=60.85447825700004, y=-43.121164516999954},
{x=61.12221914300005, y=-43.30235755099994},
{x=61.34813060100004, y=-42.60374915299991},
{x=61.78359609600005, y=-42.69566809799994},
{x=62.36123281500005, y=-42.64045162699995},
{x=62.46784088700008, y=-42.366444464999915},
{x=62.827337958000044, y=-42.55455481699994},
{x=62.92475006700005, y=-41.76024329299992},
{x=63.17096588700008, y=-41.49156653599994},
{x=63.48582591400003, y=-41.43814042899993},
{x=63.507066148000035, y=-40.875111456999946},
{x=63.74632396000004, y=-40.53282630099994},
{x=64.23037344000005, y=-41.44871985599991},
{x=64.73932526200008, y=-40.73412024599992},
{x=65.02142975500004, y=-40.30040442599994},
{x=65.37763092700004, y=-39.88267981699994},
{x=65.60541413000004, y=-39.50914466099994},
{x=65.59202708500004, y=-38.77989661399994},
{x=65.63727448100008, y=-38.24583899599992},
{x=65.94676341400003, y=-38.069203253999945},
{x=66.38507721600007, y=-37.86168372299994},
{x=65.94257233300004, y=-37.69310462099992},
{x=65.93598053600005, y=-36.740345831999946},
{x=66.07754140800006, y=-35.917307094999956},
{x=66.40387604400007, y=-35.85374915299991},
{x=66.34625885600008, y=-34.824777798999946},
{x=66.68256256700005, y=-34.27131100199995},
{x=67.02977122600004, y=-33.85562089799993},
{x=67.41290924700007, y=-33.26988684799994},
{x=67.97040436400005, y=-32.10680091099994},
{x=68.44350820500006, y=-32.31480872299994},
{x=68.13226959800005, y=-31.453724738999938},
{x=68.20384349200003, y=-30.103342251999948},
{x=68.35138580900008, y=-29.17955481699994},
{x=68.55744049700007, y=-27.527902798999946},
{x=68.88857656500005, y=-25.789214647999927},
{x=69.25633372600004, y=-24.793120897999927},
{x=69.59080638200004, y=-23.98338782499991},
{x=69.95164622600004, y=-22.91112219999991},
{x=70.31321849200003, y=-24.859445766999954},
{x=70.03900788000004, y=-27.804066535999937},
{x=70.47443268400008, y=-28.573597785999937},
{x=71.02252838700008, y=-27.789662238999938},
{x=71.48407623900005, y=-26.70571855399993},
{x=72.13100820500006, y=-28.632801886999914},
{x=71.31150950700004, y=-25.121978318999936},
{x=70.67755768400008, y=-22.63890540299991},
{x=70.59007396000004, y=-21.771148240999935},
{x=71.00421784100007, y=-21.74364173099991},
{x=71.43480052300004, y=-21.73778235599991},
{x=71.62055084800005, y=-22.155181443999936},
{x=71.84516022300005, y=-22.72874915299991},
{x=72.40200429900005, y=-24.923451300999943},
{x=72.76536692900004, y=-25.201893683999913},
{x=73.19086334800005, y=-25.887277798999946},
{x=73.44562409100007, y=-27.170765753999945},
{x=73.54144928600005, y=-24.456857876999948},
{x=73.40859609600005, y=-21.54987545499995},
{x=73.89398834800005, y=-21.788685675999943},
{x=74.16974518400008, y=-22.209828253999945},
{x=74.29633209800005, y=-20.222401495999918},
{x=74.77435944200005, y=-20.652740037999934},
{x=75.54808177300004, y=-22.467518683999913},
{x=75.68423086100006, y=-19.61229407499991},
{x=76.17084381700005, y=-20.558216925999943},
{x=76.55499909100007, y=-22.086293097999942},
{x=76.64642975500004, y=-21.623646613999938},
{x=76.74091217700004, y=-18.51121985599991},
{x=77.51137929900005, y=-20.463490363999938},
{x=77.85248444200005, y=-21.803578253999945},
{x=79.11286041900007, y=-19.52171790299991},
{x=79.35024648600006, y=-19.493072068999936},
{x=80.21800364800004, y=-17.288441535999937},
{x=80.64838288000004, y=-15.420399542999917},
{x=81.77309804900005, y=-16.14598548099991},
{x=81.55475495000007, y=-19.24632727799991},
{x=80.70734284100007, y=-23.47402910099993},
{x=82.06246572450004, y=-22.171803351499932},
{x=82.0303234491429, y=-26.415342344857056},
{x=82.05910879100004, y=-30.317209439333283},
{x=82.83828359600005, y=-23.662220831999946},
{x=82.95697663000004, y=-31.590565558999913},
{x=83.15961334800005, y=-33.43980872299994},
{x=83.61084625850006, y=-33.25108801949992},
{x=83.35952383000006, y=-37.813059048999946},
{x=82.79376862200007, y=-38.56700598899994},
{x=82.64166901200008, y=-41.82909094999991, name="END_GREENLAND"},

{x=31.13621653900009, y=33.390635613000086, name="START_AFRICA", disconnect=true},
{x=31.078192450000074, y=33.01392662900008},
{x=31.230454820000034, y=32.289805535000085},
{x=31.202866929000038, y=32.031260613000086},
{x=31.353989976000037, y=31.858409050000034},
{x=31.36286041900007, y=32.05600019600007},
{x=31.44440338700008, y=30.93726647200009},
{x=31.265285549000055, y=30.240896030000044},
{x=30.91079336100006, y=28.83570397200009},
{x=31.449652411000045, y=26.965505405000044},
{x=32.20970286700009, y=23.27320397200009},
{x=32.36082591400009, y=20.277191602000073},
{x=31.497259833000044, y=15.62281334700009},
{x=33.22410716400009, y=11.392751498000052},
{x=34.13349030200004, y=10.018402540000068},
{x=35.53375885600008, y=11.023936394000089},
{x=37.03217194200005, y=10.894704623000052},
{x=37.259019273000035, y=9.863942905000044},
{x=36.92593008000006, y=8.355967644000089},
{x=36.92031484600005, y=7.108083530000044},
{x=36.895412502000056, y=4.787119988000086},
{x=36.60618724200009, y=2.40015709700009},
{x=35.92182038000004, y=0.069956902000058},
{x=35.47573476800005, y=-1.237049933999913},
{x=35.43305084800005, y=-2.952219204999949},
{x=35.685207424000055, y=-5.304798956999946},
{x=33.732489325000074, y=-7.387115037999934},
{x=31.864488023000035, y=-9.53538977799991},
{x=29.257961330000057, y=-10.320871548999946},
{x=26.275946356000077, y=-14.353179490999935},
{x=23.959865627000056, y=-15.768137173999946},
{x=22.832098700000074, y=-16.29645748599995},
{x=21.03123607000009, y=-17.01585852799991},
{x=20.430853583000044, y=-16.347727016999954},
{x=19.41034577000005, y=-16.44245357999995},
{x=15.794663804000038, y=-16.536000128999945},
{x=14.076117255000042, y=-16.678578253999945},
{x=13.773627020000049, y=-16.710194464999915},
{x=13.48818594000005, y=-15.500396287999934},
{x=12.655666408000059, y=-15.51280676999994},
{x=12.48541901200008, y=-16.80296790299991},
{x=12.18032461100006, y=-16.26113847599993},
{x=11.571682033000059, y=-15.081532355999911},
{x=10.971136786000045, y=-15.072336391999954},
{x=11.028265692000048, y=-14.723540818999936},
{x=9.911688544000071, y=-13.702015753999945},
{x=9.272772528000075, y=-13.374501105999911},
{x=7.746161200000074, y=-12.47093665299991},
{x=7.424709377000056, y=-12.409169074999909},
{x=6.122707424000055, y=-10.344471808999913},
{x=4.51203034100007, y=-7.959868943999936},
{x=5.184759833000044, y=-5.080922003999945},
{x=5.212795315000051, y=-4.997059699999909},
{x=5.366766669000071, y=-3.134429490999935},
{x=4.834458726000094, y=-2.206288214999915},
{x=5.351507880000042, y=-0.589344855999911},
{x=6.554348049000055, y=3.411306186000047},
{x=5.724554755000042, y=5.33139082100007},
{x=5.476385809000078, y=5.533946160000085},
{x=4.475531317000048, y=5.777679884000065},
{x=4.341945705000057, y=6.22234134200005},
{x=4.453070380000042, y=6.710948113000086},
{x=4.414129950000074, y=7.010996941000087},
{x=4.718085028000075, y=7.072276238000086},
{x=4.542914130000042, y=7.559336785000085},
{x=4.776841539000088, y=8.407399936000047},
{x=4.580471096000053, y=8.724131707000083},
{x=3.992132880000042, y=9.449392123000052},
{x=3.843817450000074, y=9.723480665000068},
{x=2.09634023600006, y=9.764414910000085},
{x=1.083197333000044, y=9.657074415000068},
{x=0.683539130000042, y=9.520030144000089},
{x=0.112290757000039, y=9.586761915000068},
{x=-0.714288018999923, y=9.05054772200009},
{x=-1.209649346999925, y=8.990896030000044},
{x=-1.574883721999925, y=9.312673373000052},
{x=-1.815524997999944, y=9.246918165000068},
{x=-1.882582289999959, y=9.29273522200009},
{x=-2.554294528999947, y=9.97006269600007},
{x=-2.456801039999959, y=9.721690300000091},
{x=-5.020231727999942, y=12.07958768800006},
{x=-5.819105726999908, y=12.940928582000083},
{x=-7.55066086699992, y=12.986308166000072},
{x=-9.57740650799991, y=13.209646030000044},
{x=-11.321384372999944, y=13.80014082100007},
{x=-13.218357028999947, y=12.694102410000085},
{x=-14.468682549999926, y=12.31967207100007},
{x=-16.798520396999947, y=11.76814401100006},
{x=-20.91806405999995, y=13.459157748000052},
{x=-23.776055596999925, y=14.489512566000087},
{x=-26.389743747999944, y=15.072520379000082},
{x=-28.52434661299992, y=16.315928582000083},
{x=-30.48723723799992, y=17.37354576900009},
{x=-32.80820077899995, y=17.86817467500009},
{x=-33.558526299999926, y=18.31690514400009},
{x=-34.18499114399992, y=18.822520379000082},
{x=-34.67742278399993, y=20.18726647200009},
{x=-34.021416924999926, y=22.76889082100007},
{x=-34.01132577899995, y=24.913340691000087},
{x=-33.357354424999926, y=27.38021894600007},
{x=-31.68417734199994, y=29.405528191000087},
{x=-28.94963958099993, y=31.782725457000083},
{x=-26.114678643999923, y=32.89478600400008},
{x=-25.31804778399993, y=33.221446160000085},
{x=-23.693617445999905, y=35.360362175000034},
{x=-22.405694268999923, y=35.42798912900008},
{x=-20.988457940999922, y=35.113291863000086},
{x=-20.30055103999996, y=34.71029707100007},
{x=-19.715590101999908, y=34.705821160000085},
{x=-18.888278903999947, y=36.21143639400009},
{x=-17.93254973799992, y=36.887461785000085},
{x=-17.075372002999927, y=38.666758660000085},
{x=-15.986504815999922, y=40.134776238000086},
{x=-14.922051690999922, y=40.670909050000034},
{x=-14.26140715899993, y=40.72934004000007},
{x=-13.066013278999947, y=40.56568444100009},
{x=-12.42669036299992, y=40.514659050000034},
{x=-10.791599216999941, y=40.500010613000086},
{x=-10.25945403399993, y=40.184336785000085},
{x=-9.642836195999905, y=39.72193444100009},
{x=-8.898858330999929, y=39.388682488000086},
{x=-7.752536716999941, y=39.35523522200003},
{x=-6.731052341999941, y=39.24586022200003},
{x=-5.443291924999926, y=38.99048912900008},
{x=-4.529961846999925, y=39.43100019600007},
{x=-3.574883721999925, y=39.79127037900008},
{x=-2.341973565999922, y=40.79346764400009},
{x=-1.967054945999962, y=41.19507897200003},
{x=-0.821709893999923, y=42.135508660000085},
{x=1.097479559000078, y=44.03288821700005},
{x=7.589056708000044, y=49.744151238000086},
{x=10.425604559000078, y=51.240977410000085},
{x=11.47882721600007, y=51.12663821700005},
{x=11.462144273000092, y=49.65007571700005},
{x=10.871893622000073, y=46.865977410000085},
{x=10.589260158000059, y=44.09750410200007},
{x=11.567206122000073, y=42.51498457100007},
{x=12.884588934000078, y=42.99350019600007},
{x=13.606390692000048, y=42.20167076900009},
{x=14.537665106000077, y=41.25318444100009},
{x=14.973578192000048, y=40.39210045700008},
{x=15.349269924000055, y=39.78093509200005},
{x=15.92877838700008, y=39.29411868600005},
{x=18.204901434000078, y=38.411143425000034},
{x=19.15965403900009, y=37.33350670700008},
{x=21.102118231000077, y=37.196299675000034},
{x=22.762989315000084, y=35.84754720600006},
{x=24.121274980000067, y=35.52418917000006},
{x=26.81777578300006, y=33.94581139400009, name="END_AFRICA"},

{x=55.473089911000045, y=21.10840905000009, name="START_EURASIA1", disconnect=true},
{x=55.24868398600006, y=21.267588738000086},
{x=56.462103583000044, y=21.04468834700009},
{x=57.743841864000046, y=22.600271030000044},
{x=57.98383209800005, y=24.381114129000082},
{x=58.54486725500004, y=23.584320509000065},
{x=58.991400458000044, y=23.628103061000047},
{x=59.47060781500005, y=24.333832227000073},
{x=59.62864817900004, y=25.684825066000087},
{x=59.691107489000046, y=28.223887566000087},
{x=60.18427155200004, y=29.052012566000087},
{x=60.59511953300006, y=28.283050977000073},
{x=60.58494700700004, y=27.21461022200009},
{x=60.50812409100007, y=26.46583092500009},
{x=60.34096914300005, y=25.90723717500009},
{x=60.16266510600008, y=25.06739342500009},
{x=60.060248114000046, y=24.278575066000087},
{x=59.84796784100007, y=22.946136915000068},
{x=60.306626695000034, y=22.998057488000086},
{x=60.58783600500004, y=21.846039259000065},
{x=60.86318594000005, y=21.382334832000083},
{x=61.35496653900003, y=21.530772332000083},
{x=62.10911692900004, y=21.31226647200009},
{x=62.76703522300005, y=21.10792076900009},
{x=63.23940664300005, y=21.743662957000083},
{x=63.452337958000044, y=22.394786004000082},
{x=63.76129791900007, y=22.99000084700009},
{x=64.31655508000006, y=23.940928582000083},
{x=64.88422272300005, y=25.177256707000083},
{x=65.32684967700004, y=25.310801629000082},
{x=65.78168366100005, y=24.426117384000065},
{x=65.80292389500005, y=23.32585696700005},
{x=65.73053620000007, y=22.273285352000073},
{x=65.53180573100008, y=22.083994988000086},
{x=65.32245514500005, y=21.435557488000086},
{x=64.78119538000004, y=21.15398196700005},
{x=64.52798086100006, y=21.36695397200009},
{x=63.96556224200003, y=20.863291863000086},
{x=63.60248444200005, y=20.030528191000087},
{x=63.25307851800005, y=19.133067254000082},
{x=63.04962799700007, y=18.492930535000085},
{x=62.79901764500005, y=18.243662957000083},
{x=62.727728583000044, y=17.995371941000087},
{x=62.33266836100006, y=17.38062584700009},
{x=61.788316148000035, y=17.38835696700005},
{x=61.456284898000035, y=17.157481316000087},
{x=60.948431708000044, y=17.158213738000086},
{x=60.50226471600007, y=18.018321160000085},
{x=60.055975653000075, y=18.762543165000068},
{x=59.61676666900007, y=18.679698113000086},
{x=59.41596100500004, y=18.536631707000083},
{x=59.10809967700004, y=18.241709832000083},
{x=59.08075592700004, y=17.66578209700009},
{x=58.70286692900004, y=17.08301842500009},
{x=58.45209381700005, y=16.830332879000082},
{x=58.10415273600006, y=16.716075066000087},
{x=57.89826080900008, y=16.48015384200005},
{x=57.44896067900004, y=16.63843834700009},
{x=56.79437897300005, y=16.472504102000073},
{x=56.16596100500004, y=15.452321811000047},
{x=56.05805084800005, y=14.60328209700009},
{x=55.38751862200007, y=12.910411004000082},
{x=56.22247955900008, y=12.70484459700009},
{x=57.18414948100008, y=12.160980665000068},
{x=57.60618724200003, y=11.879649285000085},
{x=58.32709381700005, y=11.726735873000052},
{x=58.34625885600008, y=11.220876498000052},
{x=59.00112539300005, y=11.414073113000086},
{x=59.334051825000074, y=10.734711134000065},
{x=59.62335846600007, y=10.610524936000047},
{x=59.05971914300005, y=10.306895379000082},
{x=58.993394273000035, y=9.70866946700005},
{x=58.66486237200007, y=9.190196160000085},
{x=58.10346100500004, y=8.109873894000089},
{x=58.10537344000005, y=6.980316602000073},
{x=58.32371653900003, y=6.664073113000086},
{x=58.99860260600008, y=5.535655144000089},
{x=59.277044989000046, y=6.202403191000087},
{x=59.40761953300006, y=6.00953209700009},
{x=59.390285549000055, y=5.607269727000073},
{x=59.71442291900007, y=5.498708530000044},
{x=59.809475002000056, y=6.198252800000091},
{x=60.31346263200004, y=6.277354363000086},
{x=60.513006903000075, y=6.647715691000087},
{x=60.092230536000045, y=5.750661655000044},
{x=60.187201239000046, y=5.474864129000082},
{x=60.75775788000004, y=5.675629102000073},
{x=60.76386139500005, y=5.238942905000044},
{x=61.03217194200005, y=5.31226647200009},
{x=61.15033600500004, y=6.78679446700005},
{x=61.28164297100005, y=7.318207227000073},
{x=61.27875397300005, y=6.603363477000073},
{x=61.18797435100004, y=5.059336785000085},
{x=61.43073151200008, y=5.111664259000065},
{x=61.667181708000044, y=4.944102410000085},
{x=61.82005442900004, y=5.964366082000083},
{x=61.92918528900003, y=5.443614129000082},
{x=62.13580963700008, y=5.270030144000089},
{x=62.218410549000055, y=5.992035352000073},
{x=62.280462958000044, y=7.112803582000083},
{x=62.619574286000045, y=6.621592644000089},
{x=62.730902411000045, y=7.948252800000091},
{x=62.96499258000006, y=7.776133660000085},
{x=63.08026764500005, y=8.133962436000047},
{x=63.130113023000035, y=8.188487175000091},
{x=63.29401276200008, y=9.09343509200005},
{x=63.427720445000034, y=9.96062259200005},
{x=63.57404205900008, y=10.768321160000085},
{x=64.10716380400004, y=11.299978061000047},
{x=63.64411041900007, y=9.833262566000087},
{x=63.941839911000045, y=9.972666863000086},
{x=64.29413483300004, y=10.46257571700005},
{x=64.47247955900008, y=10.715098504000082},
{x=64.42015208500004, y=11.369639519000089},
{x=64.78363678600005, y=11.79037519600007},
{x=64.91836172100005, y=11.594737175000091},
{x=65.15350983300004, y=12.637217644000089},
{x=65.35932038000004, y=12.512950066000087},
{x=65.71043528900003, y=12.448008660000085},
{x=65.89134349200003, y=13.108083530000044},
{x=66.17275625200006, y=13.705088738000086},
{x=66.32782623900005, y=13.33765709700009},
{x=66.56281159100007, y=13.423350457000083},
{x=66.86371491100005, y=13.657074415000068},
{x=67.04295482000003, y=14.595876498000052},
{x=67.14203522300005, y=15.394704623000052},
{x=67.24160390800006, y=14.328298373000052},
{x=67.50165436400005, y=15.191416863000086},
{x=67.55980052300004, y=15.436371290000068},
{x=67.71991608300004, y=15.193695509000065},
{x=67.86676666900007, y=15.524261915000068},
{x=68.15277741100005, y=15.658457879000082},
{x=67.86127350500004, y=16.31023196700005},
{x=68.13226959800005, y=16.24122155000009},
{x=68.35138580900008, y=16.22006269600007},
{x=68.46869538000004, y=17.097178582000083},
{x=68.67413971600007, y=17.57667076900009},
{x=69.35252513200004, y=18.138682488000086},
{x=69.32831452000005, y=18.954600457000083},
{x=69.31842682500007, y=19.454600457000083},
{x=69.42011139500005, y=19.551931186000047},
{x=69.75706614800004, y=20.356293165000068},
{x=69.74876536700003, y=20.51335696700005},
{x=69.73956940300008, y=22.086273634000065},
{x=70.22549062700006, y=21.23568769600007},
{x=70.16754791900007, y=22.291758660000085},
{x=69.97809479400007, y=23.420583530000044},
{x=70.65753815300008, y=24.28223717500009},
{x=70.09243398600006, y=25.033376498000052},
{x=70.83539459800005, y=26.720957879000082},
{x=70.82542552300004, y=28.363291863000086},
{x=70.45148346600007, y=28.08765709700009},
{x=70.51764557500007, y=28.458832227000073},
{x=70.66469961100006, y=30.18604576900009},
{x=70.11115143400008, y=28.78646894600007},
{x=69.73753489800004, y=30.019053582000083},
{x=69.75763580900008, y=31.199554884000065},
{x=69.79189687700006, y=32.727305535000085},
{x=69.43451569200005, y=32.19361412900008},
{x=69.27000560100004, y=33.47592207100007},
{x=69.20425039300005, y=34.96078535200007},
{x=68.55414459800005, y=38.02369225400008},
{x=68.01618073100008, y=40.01978600400008},
{x=67.19037506700005, y=41.358164910000085},
{x=66.38739655200004, y=35.67107181100005},
{x=66.78974030200004, y=33.60564212300005},
{x=66.94513580900008, y=32.47242272200009},
{x=66.52240631700005, y=33.287771030000044},
{x=65.90949127800008, y=34.69507897200009},
{x=65.39179108300004, y=34.36898847700007},
{x=64.63663611700008, y=34.93609622200006},
{x=63.905523197000036, y=37.719401119000054},

{x=46.43569570500006, y=35.47575931100005, disconnect=true},
{x=46.04755280200004, y=34.838552280000044},
{x=45.25153229400007, y=36.42603600400008},
{x=44.82050202000005, y=34.75700931100005},
{x=44.729803778000075, y=33.53679446700005},
{x=45.73627350500004, y=33.16342207100007},
{x=46.10382721600007, y=33.357758009000065},
{x=46.34955475500004, y=31.837412957000083},
{x=46.712876695000034, y=31.970713738000086},
{x=46.649603583000044, y=31.899424675000034},
{x=46.51585521000004, y=30.754893425000034},
{x=45.75112539300005, y=29.850840691000087},
{x=45.453558661000045, y=29.686534050000034},
{x=44.85911692900004, y=29.101817254000082},
{x=44.68195221600007, y=28.959971550000034},
{x=43.93219635600008, y=28.640879754000082},
{x=42.21198151200008, y=27.815277540000068},
{x=41.174709377000056, y=29.069346550000034},
{x=41.050441799000055, y=28.044932488000086},
{x=40.83120351800005, y=26.00831139400009},
{x=40.95815664300005, y=24.517751498000052},

{x=46.640692450000074, y=36.23292076900009, disconnect=true},
{x=47.05841705900008, y=38.13038170700008},
{x=47.18976471600007, y=39.25603274800005},
{x=46.704291083000044, y=38.23096764400009},
{x=45.85488515800006, y=38.049815300000034},
{x=45.52708567900004, y=37.64665774800005},
{x=45.336371161000045, y=37.51579837300005},
{x=45.318019924000055, y=36.95679772200009},
{x=44.61375560100004, y=37.922618035000085},
{x=43.122259833000044, y=40.52898196700005},
{x=41.29572174700007, y=41.19076582100007},
{x=40.993353583000044, y=38.775401238000086},
{x=41.37409088700008, y=36.633555535000085},
{x=41.791449286000045, y=35.20199629000007},
{x=41.81073639500005, y=32.53484134200005},
{x=40.76349518400008, y=29.890147332000083},
{x=39.470770575000074, y=26.252289259000065},
{x=39.01512278900009, y=26.814707879000082},
{x=38.04559967700004, y=26.885590040000068},
{x=36.92593008000006, y=28.165293816000087},
{x=36.753973700000074, y=27.449554884000065},
{x=36.77130768400008, y=28.125254754000082},
{x=36.704738674000055, y=28.230316602000073},
{x=36.816066799000055, y=28.603526238000086},
{x=36.623928127000056, y=29.068207227000073},
{x=36.17792389500005, y=29.643239780000044},
{x=36.29193756700005, y=30.478851759000065},
{x=36.402411200000074, y=32.17432701900009},
{x=36.26398346600007, y=33.82178795700008},
{x=36.59442780200004, y=35.496836785000085},
{x=36.67047760600008, y=36.215586785000085},
{x=35.493475653000075, y=35.830088738000086},
{x=33.995306708000044, y=35.636729363000086},
{x=32.90582916900007, y=35.07960045700008},
{x=31.13621653900009, y=33.390635613000086},

{x=26.81777578300006, y=33.94581139400009, disconnect=true},
{x=27.73187897300005, y=33.55250084700009},
{x=28.747870184000078, y=32.820648634000065},
{x=29.972398179000038, y=32.563487175000034},
{x=29.101385809000078, y=33.061208530000044},
{x=28.10382721600007, y=33.70142662900008},
{x=27.98305898600006, y=34.42741946700005},
{x=29.164496161000045, y=34.69499759200005},
{x=29.045070705000057, y=34.852549675000034},
{x=28.04169342700004, y=34.64144941500007},
{x=28.08543528900009, y=34.920909050000034},
{x=26.91351959800005, y=36.00025475400008},
{x=25.155462958000044, y=37.250743035000085},
{x=24.07232347000007, y=38.053968085000065},
{x=22.93846263200004, y=38.905284050000034},
{x=21.92191576400006, y=38.94959847900009},
{x=20.905178127000056, y=39.34498131600009},
{x=20.21820709800005, y=40.14958743600005},
{x=19.11359284100007, y=41.08643639400009},
{x=18.204291083000044, y=41.525726759000065},
{x=17.10773346600007, y=42.420583530000044},
{x=15.516058661000045, y=42.75993899800005},
{x=14.847479559000078, y=42.894541863000086},
{x=12.906195380000042, y=43.422618035000085},
{x=12.748846747000073, y=44.830821160000085},
{x=13.436346747000073, y=46.23454837300005},
{x=14.53266022300005, y=49.137054884000065},
{x=15.56313711100006, y=52.03874759200005},
{x=16.92536041900007, y=53.99537194100009},
{x=17.475165106000077, y=55.24008222700007},
{x=19.02997467700004, y=57.839854363000086},
{x=20.441555080000057, y=58.21729576900003},
{x=21.62490469000005, y=59.44857832100007},
{x=23.318101304000038, y=58.92945397200003},
{x=24.741848049000055, y=56.48682701900003},
{x=26.27952708500004, y=56.36378014400003},
{x=25.64679596600007, y=55.73389733200008},
{x=24.79092031500005, y=54.73780358200008},
{x=24.203599351000037, y=54.18702233200008},
{x=24.152167059000078, y=52.69743899800005},
{x=24.232814846000053, y=51.66293379000007},
{x=25.720648505000042, y=51.555023634000065},
{x=25.646144924000055, y=50.977549675000034},
{x=24.913885809000078, y=50.68140709700003},
{x=26.728583075000074, y=50.069102410000085},
{x=27.331284898000035, y=49.223399285000085},
{x=27.615057684000078, y=48.884776238000086},
{x=28.023993231000077, y=48.747080925000034},
{x=29.21352475900005, y=48.10613070200009},
{x=30.372748114000046, y=49.176442905000044},
{x=29.97296784100007, y=50.13754316500007},
{x=28.83201732000009, y=50.87614993600005},
{x=27.117336330000057, y=53.01531009200005},
{x=26.96751536700009, y=55.63233483200008},
{x=26.954291083000044, y=56.940114780000044},
{x=25.573879299000055, y=58.24830162900008},
{x=25.292790076000074, y=60.45564507600005},
{x=25.107123114000046, y=62.09717858200008},

{x=69.35854726800005, y=60.445648634000065, disconnect=true},
{x=68.56928131700005, y=59.06950931100005},
{x=68.81757233300004, y=57.949880405000044},
{x=68.24583567900004, y=54.08838951900003},
{x=68.89325592700004, y=53.62281334700003},
{x=68.42401764500005, y=52.29615319100009},
{x=67.75446198100008, y=48.78142337300005},
{x=66.91937897300005, y=47.59066816500007},
{x=67.81787450900003, y=46.44372387300007},
{x=68.23208242400005, y=44.236338738000086},
{x=67.15866249100003, y=44.09553867500006},
{x=66.18618398600006, y=43.71461022200003},
{x=65.72996653900003, y=40.27979576900009},
{x=64.63971588700008, y=39.330821160000085},
{x=65.14714128400004, y=36.820001697000066},
{x=63.905523197000036, y=37.719401119000054},

{x=25.107123114000046, y=62.09717858200008, name="START_ASIA", disconnect=true},
{x=25.334906317000048, y=63.944102410000085},
{x=25.31659577000005, y=64.86109459700003},
{x=25.48786041900007, y=66.20045006600003},
{x=24.064154364000046, y=67.46729576900003},
{x=23.833644924000055, y=67.91675866000008},
{x=23.52147044500009, y=68.45866946700005},
{x=22.780503648000035, y=69.51075280000003},
{x=22.372259833000044, y=69.50269616000008},
{x=21.46580638200004, y=69.80860436300009},
{x=21.289984442000048, y=72.10775800900007},
{x=21.39008209800005, y=72.57374108200008},
{x=20.887152411000045, y=72.85425866000008},
{x=19.762640692000048, y=72.73454837300005},
{x=19.252427476000037, y=72.83008873800003},
{x=18.911363023000035, y=72.92807050900007},
{x=18.27301666900007, y=73.07154381600003},
{x=17.53937409100007, y=73.13168379000007},
{x=16.53815338700008, y=73.35084069100003},
{x=15.619574286000045, y=73.72828209700003},
{x=14.949774481000077, y=74.03101647200003},
{x=14.533758856000077, y=74.37468509200005},
{x=13.439113674000055, y=74.69076582100007},
{x=12.043158270000049, y=75.30420983200008},
{x=9.695502020000049, y=76.37712649800005},
{x=8.978989976000037, y=76.56153405000003},
{x=8.669867255000042, y=78.13672936300003},
{x=9.416449286000045, y=78.93344160200007},
{x=10.28221263200004, y=79.85132897200003},
{x=15.868068752000056, y=80.54167728000004},
{x=16.37368398600006, y=81.42408287900008},
{x=18.29828522300005, y=84.11312910200007},
{x=19.789943752000056, y=85.29029381600003},
{x=20.063666083000044, y=86.18775475400008},
{x=20.75185781500005, y=86.98658287900008},
{x=21.62400950700004, y=88.45630944100003},
{x=22.16229889500005, y=88.64429772200003},
{x=21.857123114000046, y=89.65691165500004},
{x=21.99941640800006, y=89.98519941500007},
{x=22.510158596000053, y=90.45118248800003},
{x=22.767401434000078, y=91.34644616000003},
{x=21.94489166900007, y=91.89820397200003},
{x=21.29954661700009, y=92.04525800900007},
{x=20.478949286000045, y=92.83155358200008},
{x=19.99481842700004, y=93.04867597700007},
{x=19.902899481000077, y=93.54672285200007},
{x=19.437445380000042, y=93.98438561300003},
{x=19.032049872000073, y=93.95248457100007},
{x=18.73851146000004, y=94.15642337300005},
{x=17.303290106000077, y=94.58448326900003},
{x=16.007147528000075, y=94.34799238400007},
{x=16.131008205000057, y=94.77271569100003},
{x=15.847398179000038, y=94.89633222700007},
{x=16.148382880000042, y=95.36158287900008},
{x=17.44171784100007, y=96.87159264400003},
{x=16.53815338700008, y=97.54835045700003},
{x=15.266831773000092, y=97.74317467500003},
{x=14.46873607000009, y=98.00025475400008},
{x=13.651556708000044, y=98.19336998800003},
{x=12.363959052000041, y=98.69996178500003},
{x=9.719916083000044, y=98.50049889400003},
{x=8.584906317000048, y=98.22527103000004},
{x=8.19920482000009, y=98.74740644600007},
{x=7.507147528000075, y=99.29558353000004},
{x=6.705267645000049, y=99.91749108200003},
{x=5.061428127000056, y=100.37232506600003},
{x=2.833563544000071, y=101.31055748800003},
{x=1.342840887000079, y=103.55103600400008},
{x=1.934637762000079, y=104.12232506600003},
{x=4.168972072000088, y=103.44280032600005},
{x=6.917873440000051, y=101.32829837300005},
{x=7.616400458000044, y=100.32349694100003},
{x=9.233303127000056, y=99.88599694100003},
{x=10.075832424000055, y=99.15349368600005},
{x=11.14874909100007, y=99.50131269600007},
{x=13.152777411000045, y=100.07398522200003},
{x=13.083482164000088, y=100.89210045700003},
{x=12.70213450700004, y=101.68637129000007},
{x=12.39940013200004, y=102.21461022200003},
{x=11.892279364000046, y=102.78345787900008},
{x=11.536322333000044, y=102.98462975400008},
{x=11.053127346000053, y=103.10084069100003},
{x=9.857082424000055, y=105.11329186300003},
{x=8.984767971000053, y=105.40894616000003},
{x=10.22923411700009, y=106.11784915500004},
{x=10.167669989000046, y=106.76514733200003},
{x=10.673529364000046, y=106.79354902400007},
{x=10.489569403000075, y=107.06267337300005},
{x=11.104681708000044, y=108.49170983200003},
{x=11.824042059000078, y=109.16334069100003},
{x=12.660101630000042, y=109.34205162900003},
{x=13.559759833000044, y=109.25831139400003},
{x=14.451605536000045, y=109.13021894600007},
{x=15.356390692000048, y=108.79623457100007},
{x=16.15892161700009, y=108.25269616000003},
{x=16.387518622000073, y=107.82162519600007},
{x=17.44725169500009, y=106.66358483200003},
{x=18.100490627000056, y=106.30632571700005},
{x=19.13385651200008, y=105.72982832100007},
{x=20.972357489000046, y=107.02564537900003},
{x=22.464626369000086, y=114.49114017000011},
{x=22.66904582500007, y=114.73745015700001},
{x=22.966620184000078, y=116.26937910200002},
{x=24.18626536700009, y=117.90731855600006},
{x=24.584784247000073, y=118.10336347700002},
{x=25.90135325700004, y=119.64039147200003},
{x=26.807928778000075, y=119.71200605600006},
{x=26.59332916900007, y=119.94117272200003},
{x=27.31191640800006, y=120.26596113400001},
{x=28.28436223600005, y=121.12180321300002},
{x=30.589056708000044, y=121.10238691500001},
{x=32.143744208000044, y=119.76441491000003},
{x=31.865903812000056, y=121.22786432600003},
{x=32.69765859600005, y=120.83033287900003},
{x=34.07684967700004, y=120.34823652400007},
{x=35.27130768400008, y=119.41871178500003},
{x=35.98700592700004, y=120.30526777400007},
{x=36.11766185100004, y=120.62387129000001},
{x=37.42536041900007, y=122.47120201900009},
{x=37.581244208000044, y=121.30811608200008},
{x=37.13906484600005, y=119.52450605600006},
{x=38.246161200000074, y=117.90015709700003},
{x=39.15224844000005, y=118.93189537900003},
{x=40.16348444100004, y=120.289474445},
{x=40.89329661700003, y=121.13062584700003},
{x=40.28388092700004, y=122.10816491000003},
{x=39.528631903000075, y=121.22339928500003},
{x=38.96735260600008, y=121.29184004000001},
{x=38.970770575000074, y=121.89478600400003},
{x=39.55817291900007, y=122.76913496200007},
{x=39.792629299000055, y=123.59766686300009},
{x=39.829331773000035, y=124.43002363400001},
{x=39.62400950700004, y=125.01807701900009},
{x=39.14354075700004, y=125.29330488400001},
{x=38.71112702000005, y=125.54371178500003},
{x=38.44037506700005, y=124.97266686300009},
{x=37.95844147300005, y=125.00546308700007},
{x=37.69354889500005, y=125.37265058700007},
{x=37.94354889500005, y=125.83497155000009},
{x=37.727728583000044, y=126.66879316500001},
{x=37.21629466400009, y=126.71583092500009},
{x=36.92837148600006, y=127.00033613400001},
{x=34.74916010800007, y=126.36401690800005},
{x=34.58803945500006, y=126.29786217500009},
{x=34.437567450000074, y=126.70337975400003},
{x=34.678900458000044, y=127.24594160200002},
{x=34.82713450700004, y=127.42481530000009},
{x=34.97288646000004, y=127.73658287900003},
{x=34.95994700700004, y=128.4414168630001},
{x=35.08188205300007, y=128.91841959300007},
{x=35.63609001100008, y=129.44305339200002},
{x=37.87677643400008, y=128.85914147200003},
{x=39.34756094000005, y=127.47339928500003},
{x=40.02610911700009, y=128.18238366000003},
{x=40.67365143400008, y=129.3032332690001},
{x=41.904262980000055, y=129.89674003200003},
{x=42.33577952100006, y=130.5101895030001},
{x=42.599595445000034, y=130.8896590500001},
{x=42.94196198100008, y=131.47730553500003},
{x=43.09031810100004, y=131.87094160200002},
{x=42.915676174000055, y=132.36630293100006},
{x=42.77912018400008, y=133.4443465500001},
{x=43.452460028000075, y=134.9638778000001},
{x=43.95978424700007, y=135.52987714900007},
{x=44.83226146000004, y=136.4580184250001},
{x=46.780951239000046, y=138.40975996200007},
{x=48.56818268400008, y=140.17196699300007},
{x=49.273098049000055, y=140.36402428500003},
{x=50.19554271000004, y=140.52027428500003},

{x=75.87287018400008, y=93.41895592500003, disconnect=true},
{x=75.41445547100005, y=89.00847415500004},
{x=74.98529694200005, y=87.17286217500003},
{x=74.82078685100004, y=86.09058678500003},
{x=74.30707428600005, y=86.78174889400003},
{x=74.01072825700004, y=87.36622155000003},
{x=73.02130768400008, y=86.74537194100003},
{x=73.64777252800008, y=81.52979576900003},
{x=72.80609772300005, y=80.77295983200008},
{x=71.74038320500006, y=83.45875084700003},
{x=70.35309479400007, y=83.53443444100003},
{x=70.91697825700004, y=82.87663821700005},
{x=70.57965729400007, y=82.08594811300003},
{x=71.73688385600008, y=82.41879316500007},
{x=71.96185944200005, y=78.22681725400008},
{x=71.10968659100007, y=78.23755944100003},
{x=71.15892161700003, y=77.38209069100003},
{x=71.93244049700007, y=75.25294030000003},
{x=72.67389557500007, y=74.99952233200008},
{x=70.38349030200004, y=74.07740319100003},
{x=69.04718659100007, y=76.86296634200005},
{x=67.90729401200008, y=77.47877037900008},
{x=67.55963776200008, y=78.51872806100005},
{x=68.29364655200004, y=77.16187584700003},
{x=67.79718659100007, y=74.79224694100003},
{x=66.24494049700007, y=72.04672285200007},
{x=66.72085195500006, y=70.46648196700005},
{x=66.92694733300004, y=71.42847741000008},
{x=67.55687083500004, y=72.54273522200003},
{x=69.15053945500006, y=72.47380618600005},
{x=70.61261627800008, y=72.68490644600007},
{x=72.30719635600008, y=72.88396243600005},
{x=72.59804922100005, y=68.93043053500008},
{x=71.00136953300006, y=66.69027754000007},
{x=70.05849844000005, y=67.29835045700008},
{x=69.31940338700008, y=68.02621504000007},
{x=68.42340729400007, y=68.13493899800005},
{x=69.28367747600004, y=64.90886478000004},
{x=69.35854726800005, y=60.445648634000065},

{x=71.03432851800005, y=155.51905358200008, disconnect=true},
{x=71.33779531500005, y=150.45240319100003},
{x=71.81232330900008, y=148.8339949880001},
{x=71.68203359600005, y=145.27686608200008},
{x=72.23004791900007, y=145.84253991000003},
{x=72.17544179900005, y=144.59913170700008},
{x=72.71784088700008, y=142.54281660200002},
{x=72.16429271000004, y=139.53296959700003},
{x=71.42877838700008, y=139.36296634200005},
{x=71.22492096600007, y=138.067230665},
{x=71.34760163000004, y=137.44703209700003},
{x=71.37685781500005, y=134.489105665},
{x=71.09979889500005, y=131.77076256600003},
{x=71.25299713700008, y=129.66521243600005},
{x=71.94159577000005, y=128.64210045700008},
{x=72.23065827000005, y=129.56413821700005},
{x=73.06610748900005, y=128.79664147200003},
{x=73.39142487200007, y=126.78467858200008},
{x=73.60297272300005, y=123.64551842500009},
{x=73.22687409100007, y=123.52442467500009},
{x=72.89154694200005, y=122.04330488400001},
{x=73.45189036700003, y=118.88257897200003},
{x=72.78900788000004, y=113.15870201900009},
{x=73.85748932500007, y=112.93441816500001},
{x=73.71531810100004, y=110.91871178500003},
{x=73.18964264500005, y=106.33301842500003},
{x=73.31317780200004, y=106.39503014400003},
{x=74.32566966400003, y=109.83423912900003},
{x=76.26642487200007, y=113.25928795700008},
{x=76.66547272300005, y=111.50171959700003},
{x=76.46893952000005, y=106.81983483200003},
{x=77.01365794500003, y=106.97046959700003},
{x=77.20600006700005, y=105.23747806100005},
{x=76.40058014500005, y=101.85971113400007},
{x=75.83722565300008, y=99.61475670700003},
{x=76.09080638200004, y=98.16138756600003},
{x=75.96088288000004, y=96.10515384200005},
{x=76.11701080900008, y=93.65251712300005},
{x=75.87287018400008, y=93.41895592500003},

{x=50.19554271000004, y=140.52027428500003, disconnect=true},
{x=51.42877838700008, y=140.89600670700008},
{x=52.29344310100004, y=141.27898196700005},
{x=53.042669989000046, y=141.18637129},
{x=53.79002513200004, y=140.35670006600003},
{x=53.52484772300005, y=138.28443444100003},
{x=53.859442450000074, y=137.6450301440001},
{x=54.22028229400007, y=137.23414147200003},
{x=54.39166901200008, y=136.75407962300005},
{x=55.08656485500006, y=135.53425505000007},
{x=56.054348049000055, y=137.472015821},
{x=56.81107498800009, y=138.57110428400006},
{x=58.39130280200004, y=140.9375106130001},
{x=59.34282293400008, y=145.91095243000007},
{x=59.26463450700004, y=147.77515709700003},
{x=59.52488841400003, y=149.20134524800005},
{x=59.55588838500006, y=151.32209981000005},
{x=59.09055475900004, y=151.10340522600006},
{x=59.215520509000044, y=154.438020693, name="END_ASIA"},

{x=-32.31365325299993, y=133.8950301440001, name="START_AUSTRALIA", disconnect=true},
{x=-32.084242445999905, y=128.01734459700003},
{x=-33.62558359199994, y=123.86475670700008},
{x=-34.00156015399995, y=122.22339928500003},
{x=-33.97380950299993, y=120.39429772200003},
{x=-35.000664971999925, y=116.46509850400003},
{x=-32.33318450299993, y=115.73585045700008},
{x=-29.292087497999944, y=114.91236412900003},
{x=-26.52212985099993, y=113.87028641900008},
{x=-26.196973145999948, y=113.76260514},
{x=-25.690524997999944, y=114.09750410200002},
{x=-22.307305596999925, y=114.13754316500001},
{x=-21.659763278999947, y=115.07740319100003},
{x=-20.66806405999995, y=116.95704186300009},
{x=-20.321465752999927, y=118.55193118600005},
{x=-19.520684502999927, y=121.12826582100001},
{x=-17.911309502999927, y=122.20785566500001},
{x=-17.069512627999927, y=123.26758873800009},
{x=-17.14072030999995, y=123.78891035200002},
{x=-16.162774346999925, y=123.74952233200008},
{x=-16.353122653999947, y=124.40015709700003},
{x=-15.45712655999995, y=124.64332116000003},
{x=-15.449639580999929, y=125.06169681100005},
{x=-14.380791924999926, y=125.98023522200003},
{x=-13.949395440999922, y=126.81006920700008},
{x=-13.897637627999927, y=127.25131269600001},
{x=-14.60051848799992, y=128.054047071},
{x=-15.29070403399993, y=128.29175866000003},
{x=-14.905368747999944, y=129.10230553500003},
{x=-14.408786716999941, y=129.35808353000004},
{x=-12.50904713299991, y=130.8808699880001},
{x=-12.133396091999941, y=132.75757897200003},
{x=-11.379164320999905, y=131.94809004},
{x=-11.723809502999927, y=133.4072371750001},
{x=-13.23528411299992, y=136.1938582690001},
{x=-13.64771900799991, y=136.04664147200003},
{x=-15.264418226999908, y=135.9516707690001},
{x=-16.15081145599993, y=137.43140709700003},
{x=-17.621351820999905, y=139.92530358200008},
{x=-16.06178150799991, y=141.43116295700008},
{x=-13.18726978999996, y=141.65552819100003},
{x=-12.516045830999929, y=141.7363387380001},
{x=-11.143243096999925, y=142.79664147200003},
{x=-14.65601978999996, y=144.92009524800005},
{x=-15.840590101999908, y=145.3730574880001},
{x=-16.868584893999923, y=145.9248153000001},
{x=-18.492120049999926, y=146.21371504},
{x=-19.37867603999996, y=147.45045006600003},
{x=-20.15325286299992, y=148.43344160200002},
{x=-20.901136976999908, y=149.02068118600005},
{x=-21.56959400799991, y=149.49878991000003},
{x=-22.298272393999923, y=149.92351321700005},
{x=-22.45875416499996, y=150.756195509},
{x=-23.51995208099993, y=150.85661868600005},
{x=-24.030043226999908, y=151.66374759200005},
{x=-25.27353280999995, y=152.79639733200008},
{x=-26.280857028999947, y=153.077484571},
{x=-27.76620859199994, y=153.39991295700008},
{x=-29.61809661299992, y=153.3291121750001},
{x=-31.625196178999943, y=152.836561964},
{x=-32.76604583099993, y=152.15015709700003},
{x=-33.54762135199991, y=151.30876712300005},
{x=-34.08603280999995, y=151.146820509},
{x=-35.092868747999944, y=150.701508009},
{x=-36.08277760199991, y=150.08765709700003},
{x=-37.25904713299991, y=149.9853621750001},
{x=-38.70322030999995, y=146.5747176440001},
{x=-39.14316171699994, y=146.4106551440001},
{x=-38.23886484199994, y=145.33855228000004},
{x=-37.91139088299991, y=144.817230665},
{x=-38.64381275799991, y=143.90552819100003},
{x=-38.27109140399995, y=141.86451256600003},
{x=-37.563571872999944, y=140.110199415},
{x=-35.48674895599993, y=139.014008009},
{x=-35.19540780999995, y=138.462250196},
{x=-34.97625090899993, y=137.7744246750001},
{x=-33.51123364999995, y=137.9290168880001},
{x=-32.5935197899999, y=137.76042728000004},
{x=-33.75123463299991, y=136.94654381600003},
{x=-34.80527109199994, y=135.79948978000004},
{x=-34.596368096999925, y=135.13892662900003},
{x=-33.997979424999926, y=135.25953209700003},
{x=-33.04159921699994, y=134.22144616000003},
{x=-32.31365325299993, y=133.8950301440001, name="END_AUSTRALIA"},

{x=43.68838431900008, y=-124.16171284099994, name="START_CANADA", disconnect=true},
{x=44.67182026200004, y=-124.06278756799992},
{x=45.648504864000074, y=-123.9418757919999},
{x=46.987208957000064, y=-124.1370690789999},
{x=48.371578287000034, y=-124.73025557199992},
{x=48.04975006700005, y=-122.87254798099991},
{x=50.88458893400008, y=-126.95348059799994},
{x=51.72239817900004, y=-127.87425696499992},
{x=52.04682038000004, y=-126.71068274599992},
{x=52.50413646000004, y=-127.8726293609999},
{x=52.39907461100006, y=-128.27212480399993},
{x=53.85761139500005, y=-128.9148656889999},
{x=53.87457916900007, y=-130.00633704299992},
{x=54.539048570000034, y=-130.28819739499994},
{x=55.248114325000074, y=-129.87755286399994},
{x=55.54718659100007, y=-130.14696204299992},
{x=54.803941148000035, y=-130.7888891269999},
{x=55.294989325000074, y=-130.9102270169999},
{x=55.93398672100005, y=-131.6345922519999},
{x=55.91095612200007, y=-132.0413305329999},
{x=56.814113674000055, y=-132.36949622299994},
{x=57.147650458000044, y=-133.22036699099993},
{x=57.592962958000044, y=-133.31309973899994},
{x=57.86758047100005, y=-133.5898331369999},
{x=58.367621161000045, y=-134.61123613199993},
{x=59.400824286000045, y=-135.33971106699994},
{x=58.35919830900008, y=-136.3776749339999},
{x=59.310248114000046, y=-139.13809160099993},
{x=59.710150458000044, y=-140.2542211579999},
{x=60.30267975500004, y=-144.7698868479999},
{x=60.60569896000004, y=-144.86978105399993},
{x=61.030585028000075, y=-147.52550208199992},
{x=60.868353583000044, y=-147.9085587229999},
{x=60.759751695000034, y=-148.3709610669999},
{x=59.833685614000046, y=-149.62507076699993},
{x=60.297756252000056, y=-152.41596432199992},
{x=59.08392975500004, y=-154.02196204299992},
{x=58.491115627000056, y=-153.96711178299992},
{x=58.136419989000046, y=-154.42780514199993},
{x=57.772365627000056, y=-155.52936764199993},
{x=57.27765534100007, y=-156.35118567599991},
{x=56.91453685100004, y=-156.94086666599992},
{x=56.64280833500004, y=-157.93268795499992},
{x=55.72284577000005, y=-160.06537838399993},
{x=55.361029364000046, y=-161.4791153639999},
{x=55.10297272300005, y=-162.12800045499992},
{x=54.85887278900003, y=-163.3897598949999},
{x=55.40692780200004, y=-162.62140865799992},
{x=55.804754950000074, y=-160.7642716139999},
{x=56.702826239000046, y=-159.2347305979999},
{x=57.485256252000056, y=-157.4287003249999},
{x=58.22089264500005, y=-157.45746822799993},
{x=58.70148346600007, y=-158.3358048169999},
{x=59.04409414300005, y=-158.48078365799992},
{x=58.429022528000075, y=-158.82526607999992},
{x=58.880938240000035, y=-160.1502486589999},
{x=58.68181770100006, y=-161.3012849579999},
{x=59.46336497600004, y=-161.82123775899993},
{x=60.58270905200004, y=-163.7314346999999},
{x=61.05703359600005, y=-164.95282955599993},
{x=63.22919342700004, y=-162.6966446609999},
{x=64.41128164300005, y=-161.1888321609999},
{x=64.55125560100004, y=-162.96056067599991},
{x=64.54824453300006, y=-165.87751217399992},
{x=65.25545696600005, y=-166.32357984599992},
{x=65.82330963700008, y=-167.5029190749999},
{x=66.47833893400008, y=-164.89875240799992},
{x=66.00897858300004, y=-162.68378658799992},
{x=66.19733307500007, y=-161.11143958199992},
{x=66.78974030200004, y=-162.52582760299993},
{x=66.64484284100007, y=-160.35444088399993},
{x=67.01727936400005, y=-161.7402237619999},
{x=67.83454010600008, y=-164.72651119699992},
{x=68.91425202000005, y=-164.79076087099992},
{x=70.08559804900005, y=-162.37494869699992},
{x=70.42959219000005, y=-160.62153072799993},
{x=70.33551666900007, y=-159.84349524599992},
{x=71.20449453300006, y=-156.05793209499993},
{x=70.84113190300008, y=-155.69123287699992},
{x=71.03156159100007, y=-154.6312149729999},
{x=70.87693919500003, y=-152.6940811839999},
{x=70.48114655200004, y=-150.8124893869999},
{x=70.31216054900005, y=-147.8880509109999},
{x=70.04401276200008, y=-144.06403561099992},
{x=69.97256094000005, y=-142.51256262899994},
{x=69.40664297100005, y=-138.95567786399994},
{x=68.76288483300004, y=-135.63565019399994},
{x=69.13027578300006, y=-135.47622636599993},
{x=69.48224518400008, y=-134.6413468089999},
{x=69.51703522300005, y=-133.9629613919999},
{x=70.09495677300004, y=-130.90859941299993},
{x=69.86908600500004, y=-130.1106664699999},
{x=69.53685130400004, y=-131.89415442599994},
{x=68.72199127800008, y=-133.0380753249999},
{x=69.45718008000006, y=-131.47594153599994},
{x=69.57562897300005, y=-131.2554418609999},
{x=70.10952383000006, y=-126.99079342399995},
{x=69.44814687700006, y=-125.24722245999993},
{x=69.74640534100007, y=-125.08584550699993},
{x=69.99860260600008, y=-124.94713294199994},
{x=70.07599518400008, y=-124.55980383999992},
{x=69.85968659100007, y=-124.40371660099993},
{x=69.25617096600007, y=-118.86876380099994},
{x=67.72553131700005, y=-111.98912512899994},
{x=67.92413971600007, y=-110.18016516799992},
{x=67.43402741100005, y=-108.92992102799991},
{x=67.21234772300005, y=-107.52700761599993},
{x=68.31500885600008, y=-104.92642167899993},
{x=68.00311920800004, y=-103.30532792899993},
{x=67.75360748900005, y=-102.01187089799993},
{x=67.79596588700008, y=-100.83128821499992},
{x=67.75226471600007, y=-98.89268958199995},
{x=67.59540436400005, y=-97.5579320949999},
{x=67.93406810100004, y=-98.22329667899993},
{x=68.35508860900006, y=-98.11693274599992},
{x=68.30048248900005, y=-97.07201087099992},
{x=68.29437897300005, y=-95.90684973899994},
{x=67.47125885600008, y=-96.11929277299993},
{x=67.15888092700004, y=-95.49168860599991},
{x=67.08734772300005, y=-96.37677975199995},
{x=66.92772044500003, y=-95.7630102199999},
{x=67.69635651200008, y=-95.66356360599991},
{x=68.39878978100006, y=-94.1387426419999},
{x=68.99567291900007, y=-93.90929114499994},
{x=68.87274811400005, y=-94.55357825399994},
{x=69.26561107000003, y=-93.83153235599991},
{x=69.59613678600005, y=-94.52794348899994},
{x=69.91425202000005, y=-96.08849036399994},
{x=70.63963450700004, y=-96.26732337099992},
{x=71.29535553600005, y=-95.60875403599994},
{x=71.82953522300005, y=-95.21605383999992},
{x=71.81159088700008, y=-94.38805091099994},
{x=71.14154694200005, y=-92.86343339799993},
{x=70.56578196800007, y=-92.25658118399991},
{x=70.35932038000004, y=-91.74429277299993},
{x=70.05552806200006, y=-91.99356035099993},
{x=69.66010163000004, y=-92.34044348899994},
{x=69.52326080900008, y=-91.57237708199995},
{x=69.30060455900008, y=-90.81655839799993},
{x=68.87897370000007, y=-90.43529212099992},
{x=68.30988190300008, y=-90.32412675699993},
{x=68.81749095300006, y=-89.68797766799992},
{x=68.20795319200005, y=-87.92735755099994},
{x=67.55353424700007, y=-87.8548070949999},
{x=67.31708405200004, y=-87.04645748599995},
{x=68.02118561400005, y=-86.12035071499992},
{x=68.72699616100005, y=-85.66812089799993},
{x=69.03461334800005, y=-84.80809485599991},
{x=69.40778229400007, y=-85.50829016799992},
{x=69.68150462400007, y=-83.35659745999993},
{x=69.13963450700004, y=-81.32168535099993},
{x=68.51276276200008, y=-81.62954667899993},
{x=68.17015208500004, y=-82.10191809799994},
{x=67.09446849200003, y=-81.41844641799992},
{x=66.35984935100004, y=-83.58307857999995},
{x=66.78229401200008, y=-84.26813717399995},
{x=66.93309153900003, y=-84.71996008999992},
{x=66.21373118700006, y=-83.79141191299993},
{x=66.27008698100008, y=-85.24156653599994},
{x=65.92572663000004, y=-86.30955969999991},
{x=65.32969798400006, y=-87.38076738199993},
{x=65.83091868700006, y=-89.83747311099995},
{x=65.32819245000007, y=-89.0398656889999},
{x=64.55817291900007, y=-87.59996497299994},
{x=63.97724030200004, y=-89.09870357999995},
{x=64.19061920800004, y=-89.8684789699999},
{x=63.91742584800005, y=-90.11607825399994},
{x=63.61277903900003, y=-91.05532792899993},
{x=63.993475653000075, y=-93.31338456899994},
{x=63.94993724200003, y=-93.44001217399995},
{x=63.71055735900006, y=-91.83462480399993},
{x=63.13694896000004, y=-90.6953832669999},
{x=62.69135163000004, y=-92.27562415299991},
{x=62.46861399900007, y=-92.53343665299991},
{x=62.338324286000045, y=-93.14940344999991},
{x=61.454413153000075, y=-93.97288977799991},
{x=60.83445872600004, y=-94.35777747299994},
{x=59.94354889500005, y=-94.80060787699995},
{x=58.68378327000005, y=-94.20799719999991},
{x=58.710191148000035, y=-93.4363500639999},
{x=57.43260325700004, y=-92.45083574099993},
{x=56.84251536700003, y=-89.03624426999994},
{x=55.91461823100008, y=-86.89549719999991},
{x=55.24363841400003, y=-83.72423255099994},
{x=55.16791413000004, y=-82.54759680899991},
{x=54.06504954600007, y=-82.24258378799993},
{x=52.25938548400006, y=-81.82017981699994},
{x=51.18842194200005, y=-80.74551347599993},
{x=50.99445221600007, y=-79.54588782499991},
{x=51.37449778900009, y=-78.87970943899995},
{x=52.236395575000074, y=-78.55296790299991},
{x=52.75739166900007, y=-78.85016842399995},
{x=53.29148997600004, y=-78.95201575399994},
{x=54.12518952000005, y=-79.10773678299995},
{x=54.56968821800007, y=-79.55040442599994},
{x=55.24864329600007, y=-77.87441972599993},
{x=57.46263255400004, y=-76.70197506399995},
{x=58.57721588700008, y=-78.38337154899995},
{x=58.98729075700004, y=-78.31501217399995},
{x=60.40424225500004, y=-77.74315344999991},
{x=61.27924225500004, y=-77.72679602799991},
{x=62.325425523000035, y=-78.10728919199994},
{x=62.34434642100007, y=-75.99762936099995},
{x=62.429673570000034, y=-73.90709387899994},
{x=61.82099030200004, y=-72.63158118399991},
{x=61.07721588700008, y=-70.41836503799993},
{x=60.92295970300006, y=-69.7403051419999},
{x=60.306626695000034, y=-69.75328528599994},
{x=58.890326239000046, y=-69.51130123599995},
{x=58.76386139500005, y=-69.92149817599994},
{x=58.74481842700004, y=-68.37832597599993},
{x=58.28510163000004, y=-68.23884029899995},
{x=58.464544989000046, y=-67.78831946499992},
{x=58.39293040600006, y=-67.06419837099992},
{x=58.682928778000075, y=-66.10118567599994},
{x=58.696600653000075, y=-66.03970292899993},
{x=59.10443756700005, y=-65.53652910099993},
{x=59.472764390000066, y=-65.24596106699994},
{x=59.03758372600004, y=-64.02985592399995},
{x=58.851996161000045, y=-63.29832923099991},
{x=57.94977448100008, y=-62.37734941299993},
{x=55.89862702000005, y=-61.17332923099991},
{x=55.78880442900004, y=-60.40562903599994},
{x=55.221909898000035, y=-59.755848761999914},
{x=54.84454987200007, y=-58.82359778599994},
{x=54.62608470300006, y=-57.44627844999991},
{x=54.145697333000044, y=-58.727406378999945},
{x=53.84796784100007, y=-59.79133053299995},
{x=53.55426666900007, y=-60.25365149599992},
{x=53.499701239000046, y=-59.86070716099994},
{x=54.10187409100007, y=-57.85142981699994},
{x=53.84467194200005, y=-57.144642706999946},
{x=53.73334381700005, y=-56.83250891799992},
{x=53.51251862200007, y=-56.02525794199994},
{x=53.02220286700003, y=-55.95571855399993},
{x=52.00804271000004, y=-55.86359615799995},
{x=51.48570384300007, y=-57.437611456999946},
{x=51.25568268400008, y=-58.671986456999946},
{x=50.761053778000075, y=-59.00796464799993},
{x=50.23106517100007, y=-60.65111243399991},
{x=50.20502350500004, y=-61.83934485599991},
{x=50.266831773000035, y=-63.53628495999993},
{x=50.20140208500004, y=-66.35704505099994},
{x=49.43345774900007, y=-67.26817786399994},
{x=48.89948151200008, y=-68.80284583199995},
{x=48.46515534100007, y=-70.98721269399994},
{x=47.565741278000075, y=-70.20685787699995},
{x=46.10960521000004, y=-73.02204342399995},
{x=45.16437409100007, y=-74.47305253799993},
{x=46.09349192900004, y=-73.01573645699995},
{x=47.52948639500005, y=-69.94229081899994},
{x=49.06163157800006, y=-64.49791419199994},
{x=48.43158600500004, y=-64.31236731699994},
{x=48.085272528000075, y=-66.32774817599994},
{x=47.81586334800005, y=-64.85627193899995},
{x=47.094631252000056, y=-65.36754309799994},
{x=46.32851797100005, y=-64.68342037699995},
{x=45.876613674000055, y=-63.732492641999954},
{x=45.76626211100006, y=-63.182199673999946},
{x=45.65525950700004, y=-62.44163977799991},
{x=45.30174388200004, y=-60.97061113199993},
{x=45.08429596600007, y=-61.75617428299995},
{x=44.85899485900006, y=-62.472767706999946},
{x=44.47500234600005, y=-63.728627081999946},
{x=44.16193268400008, y=-64.54918372299994},
{x=43.77086009300007, y=-65.01870683499993},
{x=43.518784898000035, y=-65.40778561099995},
{x=43.77903880400004, y=-65.86620032499991},
{x=44.02163320500006, y=-66.15583248599995},
{x=45.139553127000056, y=-64.86318925699993},
{x=45.32212962400007, y=-63.71979732999995},
{x=45.38304271000004, y=-64.92870032499991},
{x=45.81671784100007, y=-64.49054928299995},
{x=45.24437083500004, y=-65.75234941299993},
{x=45.09800853100006, y=-66.46491451699995},
{x=45.14284902600008, y=-67.16596432199992},
{x=44.65725332200009, y=-67.2038874989999},
{x=44.548041083000044, y=-67.68370520699995},
{x=44.32831452000005, y=-68.06053626199991},
{x=44.40562571800007, y=-68.52599036399994},
{x=44.43191152600008, y=-68.96544348899994},
{x=43.89756907800006, y=-69.5339656239999},
{x=43.775905666000085, y=-69.7307429679999},
{x=43.83230215100008, y=-69.87535559799994},
{x=43.45250071800007, y=-70.32164466099994},

{x=43.68838431900008, y=-124.16171284099994, disconnect=true},
{x=42.09795819000004, y=-124.35045799499994},
{x=40.84837474200009, y=-124.08446204299992},
{x=39.90801781000005, y=-123.9306962689999},
{x=38.32247308700005, y=-123.0759672499999},
{x=37.86081040500005, y=-122.58834125899995},
{x=38.11322666600006, y=-122.06579728499992},
{x=38.057537194000076, y=-121.5559221979999},
{x=37.907674614000086, y=-122.39108785399992},
{x=37.57469864700005, y=-122.28326872499991},
{x=37.17401999300006, y=-122.36935319999992},
{x=36.56885441600008, y=-121.97007538999992},
{x=35.88673133800006, y=-121.46320087599992},
{x=34.87140534100007, y=-120.63255774599992},
{x=34.413902574000076, y=-119.56800921999991},
{x=34.04785799500007, y=-118.95403653799991},
{x=33.76911041900007, y=-118.23070227799991},
{x=32.52411934300005, y=-117.12414985399994},
{x=30.956773179000038, y=-116.24974524599992},
{x=29.428615627000056, y=-115.18712317599994},
{x=28.307129428000053, y=-114.09929054199995},
{x=27.80831244500007, y=-115.0495085419999},
{x=26.983500183000046, y=-113.87866020799993},
{x=26.674994208000044, y=-113.07689368399991},
{x=25.471117378000088, y=-112.07937260199992},
{x=24.75346589800006, y=-112.04305371299989},
{x=24.339625712000043, y=-111.44494507399992},
{x=23.125282931000072, y=-109.51617724399995},
{x=24.062799557000062, y=-109.82828804299993},
{x=24.106242929000075, y=-110.39269389099991},
{x=24.84991078900009, y=-110.6787565429999},
{x=25.35941790800007, y=-110.98229971199993},
{x=25.90843333500004, y=-111.33901933499993},
{x=26.884027951000064, y=-111.79509108899992},
{x=27.142147339000076, y=-112.11453437699993},
{x=27.997756250000066, y=-112.76036220299994},
{x=28.819009772000072, y=-113.29100553399991},
{x=29.252714105000052, y=-113.62307566099992},
{x=29.77042683800005, y=-114.27865029399993},
{x=30.774816066000085, y=-114.7005817999999},
{x=31.717235522000067, y=-114.5406792949999},
{x=31.203138636000062, y=-113.12847332299992},
{x=29.46279531500005, y=-112.37783769399994},
{x=28.393784898000035, y=-111.51622473899994},
{x=27.898382880000042, y=-110.67096920499995},
{x=27.081692415000077, y=-110.06232888199992},
{x=26.358710028000075, y=-109.17015540299991},
{x=25.762884833000044, y=-109.35948645699995},
{x=25.58466217700004, y=-109.05776933499993},
{x=25.14524974200009, y=-108.3992406889999},
{x=24.82884349200009, y=-108.06541907499991},
{x=24.514960028000075, y=-107.56090247299994},
{x=23.294989325000074, y=-106.4707332019999},
{x=21.03066640800006, y=-105.30968176999994},
{x=19.56107250100007, y=-105.08771005999995},
{x=18.547985243000085, y=-103.62902516299994},
{x=17.19875722900008, y=-100.83389238199993},
{x=16.327148308000062, y=-98.5674605559999},
{x=15.703500897000083, y=-96.20423237099993},
{x=16.251743943000065, y=-94.90206175599991},
{x=15.092596747000073, y=-92.73741614499994},
{x=13.274969794000071, y=-88.71043860599991},
{x=13.386013088000084, y=-87.66567949099993},
{x=13.001247463000084, y=-87.3008113269999},
{x=12.45962148600006, y=-87.16209876199991},
{x=10.940578518000052, y=-85.76113847599993},
{x=10.542995510000083, y=-85.77513587099992},
{x=9.699408270000049, y=-84.99742591099994},
{x=10.136135158000059, y=-85.04605058499993},
{x=9.385321356000077, y=-84.15225175699993},
{x=8.438259182000081, y=-83.54417883999992},
{x=8.602687893000052, y=-83.18504798099991},
{x=8.301052151000079, y=-82.2424617179999},
{x=8.071600653000075, y=-81.68541419199994},
{x=7.650661526000079, y=-81.29116777299993},
{x=7.764797268000052, y=-81.07669023299991},
{x=7.265448309000078, y=-80.4284968739999},
{x=8.88226959800005, y=-79.57848059799994},

{x=9.30414459800005, y=-79.92943274599992, disconnect=true}, 
{x=8.939195054000038, y=-80.78355872299994},
{x=8.971096096000053, y=-81.91246497299994},
{x=9.20726146000004, y=-82.35057532499991},
{x=10.938666083000044, y=-83.68199622299994},
{x=12.051947333000044, y=-83.73936926999994},
{x=12.81907786700009, y=-83.59080969999991},
{x=14.776556708000044, y=-83.3450414699999},
{x=15.278143622000073, y=-83.81566321499992},
{x=15.853211763000047, y=-84.6684806159999},
{x=15.819525458000044, y=-87.41714433499993},
{x=15.75389232000009, y=-88.5781957669999},
{x=16.285549221000053, y=-88.53661048099991},
{x=18.374823309000078, y=-88.31749426999994},
{x=19.22093333500004, y=-87.63898678299995},
{x=19.567572333000044, y=-87.68325761599993},
{x=20.326971747000073, y=-87.35391191299993},
{x=21.54824453300006, y=-87.33678137899994},
{x=21.57379791900007, y=-88.23509680899991},
{x=20.23460521000004, y=-90.48729407499991},
{x=18.62181224200009, y=-91.30174719999991},
{x=18.543036200000074, y=-91.94912675699993},
{x=18.428859768000052, y=-93.13760331899994},
{x=18.232245184000078, y=-94.08458411399994},
{x=18.82453034100007, y=-95.80825761599993},
{x=21.50104401200008, y=-97.57762610599991},
{x=22.596096096000053, y=-97.89098059799994},
{x=26.019191799000055, y=-97.14073645699995},
{x=28.15375397300005, y=-97.00649980399993},
{x=28.510646877000056, y=-96.49323482999995},
{x=28.99551015800006, y=-95.23151607999995},
{x=29.96824778900009, y=-93.85802161399994},
{x=29.736476955000057, y=-92.14899654899995},
{x=29.598863023000035, y=-91.2200414699999},
{x=29.539129950000074, y=-90.13316809799994},
{x=29.18195221600007, y=-89.32176673099991},
{x=29.180080471000053, y=-89.03189042899993},
{x=29.397772528000075, y=-89.5398656889999},
{x=29.829250393000052, y=-89.45677649599992},
{x=30.052435614000046, y=-89.43472245999993},
{x=30.40375397300005, y=-88.35533606699994},
{x=30.404689846000053, y=-87.82127844999991},
{x=30.42446523600006, y=-86.3054906889999},
{x=30.11391836100006, y=-85.56969153599994},
{x=29.686346747000073, y=-85.36583411399994},
{x=30.070746161000045, y=-84.28368079299992},
{x=29.277777411000045, y=-83.09902910099993},
{x=28.400783596000053, y=-82.68472245999993},
{x=27.502101955000057, y=-82.61050370999993},
{x=26.78579336100006, y=-82.16559811099995},
{x=25.264227606000077, y=-81.06175696499992},
{x=25.173163153000075, y=-80.45091712099992},
{x=26.906683661000045, y=-80.06432044199994},
{x=28.379339911000045, y=-80.61790930899991},
{x=30.52765534100007, y=-81.47492428299995},
{x=31.133612372000073, y=-81.48729407499991},
{x=32.54018789300005, y=-80.52501380099994},
{x=32.77602773600006, y=-79.87083899599992},
{x=33.35911692900004, y=-79.27708899599992},
{x=34.207017320000034, y=-77.83267167899993},
{x=34.676988023000035, y=-77.09325110599991},
{x=34.94057851800005, y=-76.46182206899994},
{x=35.17804596600007, y=-76.5716039699999},
{x=35.44773997600004, y=-76.59919186099995},
{x=35.38418203300006, y=-76.06509355399993},
{x=35.93158600500004, y=-75.82485917899993},
{x=35.93496328300006, y=-76.37303626199991},
{x=36.14199453300006, y=-76.34943600199995},
{x=36.219631252000056, y=-75.90807044199994},
{x=36.60447825700004, y=-75.88402258999992},
{x=36.894598700000074, y=-76.31525631399995},
{x=37.31635163000004, y=-77.07249915299991},
{x=37.22242394500006, y=-76.43229875999992},
{x=37.501288153000075, y=-76.32266191299993},
{x=38.007308822000084, y=-76.89573717999991},
{x=38.00096263200004, y=-76.46914628799993},
{x=38.54901764500005, y=-77.28477942599994},
{x=38.70333104400004, y=-77.0306695239999},
{x=38.398871161000045, y=-76.88068600199995},
{x=38.42421509500008, y=-76.56164028499995},
{x=39.064003322000076, y=-76.47632745899995},
{x=39.357123114000046, y=-76.29165605399993},
{x=39.307847398000035, y=-76.17861894399994},
{x=38.92495351800005, y=-76.14834550699993},
{x=38.606201199000054, y=-76.05934748699991},
{x=38.23932526200008, y=-76.04881751199991},
{x=38.11082591400009, y=-75.74299068899995},
{x=37.59487539300005, y=-75.91104081899994},
{x=37.875113328000054, y=-75.42489525199994},
{x=38.38776276200008, y=-75.05113684799994},
{x=39.403390924000064, y=-75.52789117599991},
{x=39.514105536000045, y=-75.5215551419999},
{x=39.11676666900007, y=-74.73672441299993},
{x=39.80656409000005, y=-74.16586523199993},
{x=40.71789346700007, y=-74.11243542099993},
{x=41.00873800900007, y=-73.61936733399995},
{x=41.27875397300005, y=-72.58189856699994},
{x=41.68301015800006, y=-71.39146887899994},
{x=41.55072663000004, y=-70.92516028599994},
{x=41.79706452000005, y=-69.97984778599994},
{x=41.957098700000074, y=-70.65493730399993},
{x=42.376613674000055, y=-70.98363196499992},
{x=42.78038157800006, y=-70.80435136599993},
{x=43.45250071800007, y=-70.32164466099994, name="END_CANADA"},

{x=9.30414459800005, y=-79.92943274599992, name="START_SOUTH_AMERICA", disconnect=true},
{x=9.549403746000053, y=-79.17500364499995},
{x=9.290058988000055, y=-78.17665408399995},
{x=8.601385809000078, y=-77.31712805899991},
{x=7.940497137000079, y=-76.74714107999995},
{x=9.294745184000078, y=-76.11290442599994},
{x=10.230617580000057, y=-75.53551184799994},
{x=10.64329661700009, y=-75.42418372299994},
{x=10.965765692000048, y=-74.50576738199993},
{x=11.226304429000038, y=-74.23224850199995},
{x=11.826117255000042, y=-72.39199785099993},
{x=12.367621161000045, y=-71.74404863199993},
{x=11.93695709800005, y=-71.25413977799991},
{x=10.391506252000056, y=-71.72972571499992},
{x=9.268866278000075, y=-71.77806555899991},
{x=9.980617580000057, y=-71.20693925699993},
{x=11.287597019000088, y=-70.43394834899993},
{x=11.67523834800005, y=-69.82502193899995},
{x=11.540187893000052, y=-69.70030676999994},
{x=10.57843659100007, y=-68.24233964799993},
{x=10.18545156500005, y=-65.58043372299994},
{x=10.343695380000042, y=-64.38060462099992},
{x=10.62018463700008, y=-64.15721594999991},
{x=10.415187893000052, y=-62.96401933499993},
{x=10.06932200700004, y=-62.89411373599995},
{x=9.932318427000041, y=-62.423491990999935},
{x=9.596380927000041, y=-61.74331620999993},
{x=9.521307684000078, y=-60.955067511999914},
{x=8.857123114000046, y=-61.11204993399991},
{x=8.564886786000045, y=-60.987212693999936},
{x=7.629295072000048, y=-58.72821716899995},
{x=6.448377075000053, y=-57.618175109999925},
{x=6.004262377000089, y=-56.92884361699993},
{x=5.882882815000073, y=-55.063423422999904},
{x=5.540594794000071, y=-54.013010219999956},
{x=4.859320380000042, y=-52.27651933499993},
{x=4.233817846000079, y=-51.585346511999944},
{x=3.00549067500009, y=-50.997278639999934},
{x=2.119370835000041, y=-50.72061113199993},
{x=0.916489976000094, y=-50.17039954299992},
{x=-0.759698174999926, y=-51.700510219999956},
{x=-1.673923434999949, y=-52.19896399599992},
{x=-1.376397393999923, y=-50.823475714999915},
{x=-2.062920830999929, y=-50.428578253999945},
{x=-1.895114841999941, y=-49.65477454299992},
{x=-1.108819268999923, y=-48.23460852799991},
{x=-0.635407492999946, y=-47.578099591999944},
{x=-0.647067966999941, y=-47.25080318899995},
{x=-0.836602471999925, y=-47.07656816299993},
{x=-0.778415622999944, y=-46.81822669199994},
{x=-0.95095041299993, y=-46.56456228499991},
{x=-1.006605726999908, y=-46.20571855399993},
{x=-1.146091403999947, y=-45.96080481699994},
{x=-1.229359376999923, y=-45.6754049889999},
{x=-1.367445570999962, y=-45.46825110599991},
{x=-1.504571221999925, y=-44.88536536399994},
{x=-2.402113539999959, y=-44.395375128999945},
{x=-3.251722914999959, y=-44.78115800699993},
{x=-2.551364841999941, y=-44.06554114499994},
{x=-2.75904713299991, y=-44.08897864499994},
{x=-2.52507903399993, y=-43.61733964799993},
{x=-2.676446221999925, y=-42.56558183499993},
{x=-2.86492278399993, y=-41.688343878999945},
{x=-2.88250090899993, y=-40.83543860599991},
{x=-3.377048434999949, y=-39.08055579299992},
{x=-5.07545338299991, y=-36.69611568899995},
{x=-6.185804945999905, y=-35.09284420499995},
{x=-7.192966403999947, y=-34.79674231699994},
{x=-8.09693775799991, y=-34.88353430899991},
{x=-9.659274997999944, y=-35.90843665299991},
{x=-11.070896091999941, y=-37.24461829299992},
{x=-11.38876718499995, y=-37.43626868399991},
{x=-12.713636976999908, y=-38.60594641799992},
{x=-12.965427341999941, y=-38.755848761999914},
{x=-13.805352471999925, y=-39.03734290299991},
{x=-13.986260674999926, y=-38.97142493399991},
{x=-16.081475518999923, y=-38.94786536399994},
{x=-20.125258070999905, y=-40.17540442599994},
{x=-21.190463834999946, y=-40.92234944899991},
{x=-22.944512627999927, y=-42.022694464999915},
{x=-22.857598565999922, y=-43.20445716099994},
{x=-22.93954843499995, y=-44.05622311099995},
{x=-23.24863046699994, y=-44.68000240799995},
{x=-23.52906666499996, y=-45.14549719999991},
{x=-23.91570403399993, y=-46.197377081999946},
{x=-24.406182549999926, y=-47.05101477799991},
{x=-26.556084893999923, y=-48.66445878799993},
{x=-27.37664153399993, y=-48.57144120999993},
{x=-28.43173593499995, y=-48.807525193999936},
{x=-32.0095354149999, y=-52.04287675699993},
{x=-30.896091403999947, y=-50.93932044199994},
{x=-30.317803643999923, y=-51.026275193999936},
{x=-30.38534921699994, y=-51.10041256399995},
{x=-31.429131768999923, y=-52.001616990999935},
{x=-34.64470794099992, y=-54.32258053299995},
{x=-34.75896575299993, y=-56.57359778599994},
{x=-33.68303801899992, y=-58.549387173999946},
{x=-35.07053410699996, y=-57.446698158999936},
{x=-38.558851820999905, y=-58.56700598899994},
{x=-39.19410572699991, y=-62.3391820949999},
{x=-40.05413176899992, y=-62.36632239499994},
{x=-41.16090260199991, y=-63.62181555899991},
{x=-41.79338958099993, y=-64.99990800699993},
{x=-42.24187590899993, y=-63.65274003799993},
{x=-42.961195570999905, y=-64.46157792899993},
{x=-44.334666668999944, y=-65.24619496899993},
{x=-44.58651621799993, y=-65.39176024699992},
{x=-45.00321645999992, y=-65.84159184399994},
{x=-45.96751579699992, y=-67.55757025799994},
{x=-47.85947030999995, y=-66.27391516799992},
{x=-48.54830711199992, y=-66.8578115009999},
{x=-50.124647109999955, y=-68.25836740999995},
{x=-50.28867542499995, y=-68.76732952099991},
{x=-51.62986557799991, y=-69.05538131999992},
{x=-52.659927067999945, y=-70.52587829099991},
{x=-53.081963799999926, y=-72.52017167899993},
{x=-53.47909921699994, y=-72.65086829299992},
{x=-52.18987395599993, y=-73.51085364499994},
{x=-51.54265715899993, y=-72.92625891799992},
{x=-51.817071221999925, y=-72.73790442599994},
{x=-51.869398695999905, y=-72.93761145699995},
{x=-51.91155364399992, y=-73.47557532499991},
{x=-50.69841887799993, y=-73.82298743399991},
{x=-49.9353166649999, y=-74.34943600199995},
{x=-48.1989885399999, y=-73.52961178299995},
{x=-47.21819426899992, y=-74.22716223899994},
{x=-45.97193775799991, y=-74.42027747299994},
{x=-46.53378671699994, y=-73.74185136599993},
{x=-45.660088799999926, y=-73.19749915299991},
{x=-45.40097421699994, y=-72.98151607999995},
{x=-44.391289971999925, y=-72.63011633999992},
{x=-43.90203215899993, y=-73.06346594999991},
{x=-43.21493905999995, y=-72.9063207669999},
{x=-42.458916924999926, y=-72.44033769399994},
{x=-41.702243747999944, y=-72.46076412699995},
{x=-41.61052825299993, y=-73.09410559799994},
{x=-41.39584873199993, y=-73.84318468399994},
{x=-39.999932549999926, y=-73.70026607999995},
{x=-38.76140715899993, y=-73.44505774599992},
{x=-37.19150155999995, y=-73.23696855399993},
{x=-36.04973723799992, y=-72.79674231699994},
{x=-33.97283294099992, y=-71.90526282499991},
{x=-30.238213799999926, y=-71.63463294199994},
{x=-29.295586846999925, y=-71.36091061099995},
{x=-27.83294036299992, y=-71.10020911399994},
{x=-22.790622653999947, y=-70.3125707669999},
{x=-21.25505417399995, y=-70.08683179399992},
{x=-20.031426690999922, y=-70.13475501199991},
{x=-18.316415519999907, y=-70.4197391209999},
{x=-16.895196221999925, y=-72.26895097599993},
{x=-15.538050754999915, y=-74.86864492699993},
{x=-14.37314218499995, y=-76.09215247299994},
{x=-13.287286065999922, y=-76.27428137899994},
{x=-11.430759372999944, y=-77.42080644399994},
{x=-10.22884445099993, y=-78.09073175499992},
{x=-9.183727466999926, y=-78.55856518799993},
{x=-7.693634194999959, y=-79.43684103299995},
{x=-5.087299027999961, y=-81.17140029099994},
{x=-3.320733330999929, y=-80.30687415299991},
{x=-2.045179945999962, y=-79.80589758999992},
{x=-2.618259372999944, y=-80.25031490799995},
{x=-1.491621345999931, y=-80.81152079499992},
{x=-0.291919569999948, y=-80.40146946499993},
{x=1.003607489000046, y=-79.65575110599991},
{x=1.543361721000053, y=-78.85334225199995},
{x=2.306586005000042, y=-78.63695227799991},
{x=2.97797272300005, y=-77.63109290299991},
{x=3.486558335000041, y=-77.30028235599991},
{x=5.14915599200009, y=-77.37287350199995},
{x=5.888251044000071, y=-77.30239824099993},
{x=6.853989976000094, y=-77.68736731699994},
{x=7.688462632000039, y=-78.28213456899994},
{x=8.23533763200004, y=-77.98265540299991},
{x=8.393947658000059, y=-78.20376542899993},
{x=8.440130927000041, y=-78.4680883449999},
{x=9.12400950700004, y=-79.04625403599994},
{x=8.88226959800005, y=-79.57848059799994, name="END_SOUTH_AMERICA"},

{x=67.54328034100007, y=-72.42609615799995, name="START_CANADA_ISLAND", disconnect=true},
{x=68.92112864800004, y=-74.67357337099992},
{x=69.68773021000004, y=-76.40156002499992},
{x=69.82489655200004, y=-77.07258053299995},
{x=70.20652903900003, y=-78.14704342399995},
{x=70.44855377800008, y=-79.31684322799993},
{x=70.01089101800005, y=-81.3997289699999},
{x=70.20945872600004, y=-82.74262447799993},
{x=70.00495026200008, y=-83.14549719999991},
{x=70.52383047100005, y=-86.39281165299991},
{x=71.45156484600005, y=-89.98863684799994},
{x=72.18284739800004, y=-89.68122311099995},
{x=73.29266998900005, y=-88.74958248599995},
{x=71.40326569200005, y=-84.92320716099994},
{x=72.05194733300004, y=-85.74071204299992},
{x=72.04726797100005, y=-84.33971106699994},
{x=72.37787506700005, y=-85.1367895169999},
{x=73.01154205900008, y=-85.1719457669999},
{x=73.30149974200003, y=-83.60517330599993},
{x=72.06928131700005, y=-80.58100338399993},
{x=72.30341217700004, y=-79.49054928299995},
{x=72.07587311400005, y=-76.0684301419999},
{x=72.08295319200005, y=-75.13658606699994},
{x=71.66608307500007, y=-74.84284420499995},
{x=71.40900299700007, y=-73.40929114499994},
{x=70.92267487200007, y=-71.28400631399995},
{x=70.58661530200004, y=-71.45160885299993},
{x=70.04401276200008, y=-69.87690182199992},
{x=69.02659739800004, y=-68.92719479099992},
{x=68.65257396000004, y=-68.72651119699992},
{x=68.42816803600005, y=-66.72834225199995},
{x=68.13178131700005, y=-66.74925696499992},
{x=67.87132396000004, y=-66.34618079299992},
{x=67.88271719000005, y=-65.79303951699995},
{x=67.94855377800008, y=-65.01427161399994},
{x=67.82998281500005, y=-64.74079342399995},
{x=67.36273834800005, y=-64.76227779899995},
{x=66.89272695500006, y=-62.50470943899995},
{x=66.44440338700008, y=-61.824208136999914},
{x=66.02301666900007, y=-62.41087805899991},
{x=65.86298248900005, y=-62.76911373599995},
{x=65.10407135600008, y=-63.88044186099995},
{x=65.40558502800008, y=-64.29259192599994},
{x=65.35101959800005, y=-64.79222571499992},
{x=65.63837311400005, y=-64.83006751199991},
{x=66.00531647300005, y=-64.90965735599991},
{x=65.97089264500005, y=-65.75169837099992},
{x=66.24384186400005, y=-66.23106848899994},
{x=66.41917552300004, y=-67.35187740799995},
{x=66.37685781500005, y=-67.78368079299992},
{x=65.92755768400008, y=-67.65827389199995},
{x=65.48749420800004, y=-68.03343665299991},
{x=65.11570872600004, y=-66.75877844999991},
{x=64.96893952000005, y=-66.48017330599993},
{x=64.48533763200004, y=-65.51484127499992},
{x=63.296128648000035, y=-64.72040768099993},
{x=63.359808661000045, y=-64.94648189999992},
{x=63.49298737200007, y=-67.53998775899993},
{x=63.664048570000034, y=-68.40033932199992},
{x=62.973089911000045, y=-67.56794186099995},
{x=61.937567450000074, y=-66.66242428299995},
{x=62.284857489000046, y=-68.65668697799993},
{x=62.706732489000046, y=-69.53246008999992},
{x=62.92796458500004, y=-70.86790930899991},
{x=63.05565013200004, y=-71.19953365799995},
{x=63.450832424000055, y=-72.10065670499995},
{x=63.71979401200008, y=-71.58320064999992},
{x=63.740668036000045, y=-72.07616126199991},
{x=63.86713288000004, y=-72.59809322799993},
{x=64.55975983300004, y=-75.8281957669999},
{x=64.30141836100006, y=-76.18805904899995},
{x=64.32713450700004, y=-77.42100989499994},
{x=64.73159414300005, y=-78.18871008999992},
{x=65.41168854400007, y=-76.96637936099995},
{x=64.77179596600007, y=-75.44969641799992},
{x=65.08600495000007, y=-75.43700110599991},
{x=65.38202545800004, y=-74.68854732999995},
{x=65.82587311400005, y=-73.89911861899992},
{x=67.54328034100007, y=-72.42609615799995, name="END_CANADA_ISLAND"},

{x=41.429388739000046, y=54.17497806100005, name="START_ISLAND1", disconnect=true},
{x=42.14545319200005, y=52.428396030000044},
{x=45.37164948100008, y=51.957774285000085},
{x=46.86188385600008, y=52.18287194100009},
{x=46.968003648000035, y=50.757823113000086},
{x=45.74103424700007, y=47.59107506600009},
{x=43.643744208000044, y=47.47779381600009},
{x=42.39297109600005, y=48.02369225400008},
{x=40.58746979400007, y=49.938161655000044},
{x=36.954331773000035, y=50.65870201900003},
{x=38.503566799000055, y=53.82829837300005},
{x=39.92487213700008, y=53.53728274800005},
{x=40.82363515800006, y=54.27857506600009},
{x=41.429388739000046, y=54.17497806100005, name="END_ISLAND1"},

{x=54.69989655200004, y=-130.36449133999992, name="START_ISLANDS1", disconnect=true},
{x=52.39988841400003, y=-131.35521399599992},
{x=52.95026276200008, y=-132.32925370999993},
{x=53.881293036000045, y=-132.1142471999999},
{x=54.00381094000005, y=-133.09532630099994},
{x=53.43231842700004, y=-132.64122473899994},
{x=52.60252513200004, y=-128.2749731109999},
{x=53.682847398000035, y=-128.8574926419999},
{x=53.607855536000045, y=-130.05752519399994},
{x=53.39836629800004, y=-129.40279714399992},
{x=54.29677969000005, y=-130.3542374339999},
{x=53.15371328300006, y=-129.33844967399995},
{x=52.280910549000055, y=-127.61896725199989},
{x=52.322699286000045, y=-128.7280981109999},
{x=52.00234609600005, y=-127.96056067599994},
{x=53.90273672100005, y=-130.30089270699995},
{x=52.15607330900008, y=-131.0987035799999},
{x=52.753159898000035, y=-129.1357315749999},
{x=52.84739817900004, y=-128.70230058499993},
{x=53.03587474200003, y=-129.5703018869999},
{x=54.69989655200004, y=-130.36449133999992},

{x=68.68618398600006, y=-114.16657467399995, disconnect=true},
{x=73.66351959800005, y=-115.60985266799992},
{x=74.04319896000004, y=-119.7952367829999},
{x=73.99217357000003, y=-124.34227454299992},
{x=73.07550690300008, y=-124.86538652299993},
{x=72.43886953300006, y=-125.36314856699994},
{x=71.39813873900005, y=-123.5567113919999},
{x=72.22955963700008, y=-119.77281653599994},
{x=68.68618398600006, y=-114.16657467399995},

{x=70.28461334800005, y=-101.63084876199991, disconnect=true},
{x=70.97288646000004, y=-104.3040258449999},
{x=72.31264883000006, y=-105.1270238919999},
{x=73.03229401200008, y=-105.82848059799994},
{x=73.21255117400005, y=-107.90453040299991},
{x=71.77285390800006, y=-108.23924719999991},
{x=72.50438060100004, y=-110.60334225199995},
{x=72.29515208500004, y=-111.6507055329999},
{x=72.75478750200006, y=-113.55858313699991},
{x=73.01732005400004, y=-114.01646887899994},
{x=72.08161041900007, y=-118.70982825399994},
{x=71.57217031500005, y=-117.65371660099993},
{x=71.53864166900007, y=-115.4546606109999},
{x=70.60830312700006, y=-116.68285071499992},
{x=70.41453685100004, y=-112.06940670499995},
{x=69.76410553600005, y=-117.25206458199995},
{x=69.11920807500007, y=-113.66584225199995},
{x=68.56232330900008, y=-111.37100175699993},
{x=69.43325429900005, y=-106.74425208199995},
{x=68.95473867400005, y=-102.21410071499992},
{x=69.48969147300005, y=-102.34569251199991},
{x=68.55068594000005, y=-110.83604895699995},
{x=68.09813060100004, y=-109.45132402299993},
{x=68.28994375200006, y=-111.5991918609999},
{x=67.10443756700005, y=-107.32046464799993},
{x=67.95734284100007, y=-109.18468176999994},
{x=68.55438873900005, y=-104.57567298099991},
{x=69.50023021000004, y=-101.0391332669999},
{x=69.21954987200007, y=-101.5637914699999},
{x=69.61981842700004, y=-95.62950598899994},
{x=69.46727122600004, y=-96.10130774599992},
{x=69.23525625200006, y=-96.06773841099994},
{x=69.85285065300008, y=-97.70665442599994},
{x=69.27342357000003, y=-98.69871985599991},
{x=68.84357330900008, y=-98.87413489499994},
{x=68.61884186400005, y=-96.01276607999995},
{x=68.92890045800004, y=-100.22862708199995},
{x=70.28461334800005, y=-101.63084876199991},

{x=55.82440827000005, y=-133.57896887899994, disconnect=true},
{x=57.56313711100006, y=-135.61611894399994},
{x=58.04010651200008, y=-136.54845130099994},
{x=56.692572333000044, y=-134.97191321499992},
{x=55.941392320000034, y=-132.21182206899994},
{x=56.59227122600004, y=-132.95449785099993},
{x=57.05516185100004, y=-134.01134192599994},
{x=55.16791413000004, y=-132.55833899599992},
{x=55.82440827000005, y=-133.57896887899994, name="END_ISLANDS1"},

{x=59.215520509000044, y=154.438020693, name="START_RUSSIA1", disconnect=true},
{x=59.420881577000046, y=154.4268503730001},
{x=59.94630657200008, y=154.48758987500003},
{x=61.00156668200003, y=156.09117289000005},
{x=61.832571243000075, y=158.73608489700007},
{x=61.918878290000066, y=160.24080421300005},
{x=61.292124612000066, y=159.87290751600005},
{x=60.89265822100003, y=160.26813237300007},
{x=61.03569603600005, y=161.21507002400006},
{x=61.671942450000074, y=162.4072371750001},
{x=61.954659390000074, y=163.08017639800005},
{x=62.65717194200005, y=164.6162215500001},
{x=61.924335564000046, y=164.09488332},
{x=60.750624369000036, y=163.31531905600002},
{x=59.667181708000044, y=160.926117384},
{x=58.35765208500004, y=158.89283287900003},
{x=57.05425134200004, y=156.53077354200002},
{x=53.042669989000046, y=156.09791100400003},
{x=50.88260556800009, y=156.658235246},
{x=52.206691799000055, y=158.38648522200003},
{x=52.82330963700008, y=158.592133009},
{x=53.19037506700005, y=159.79957116000003},
{x=53.68744538000004, y=159.8525496750001},
{x=55.13070359500006, y=161.84264852400008},
{x=56.358756190000065, y=162.87251276200004},
{x=56.72235241800007, y=162.834629844},
{x=58.20620273800006, y=162.04692665000005},
{x=59.15892161700003, y=163.1718856130001},
{x=59.99750397300005, y=163.7656356130001},
{x=59.86676666900007, y=164.78296959700003},
{x=60.36562734600005, y=165.94060306100005},
{x=60.41648997600004, y=167.28150475400003},
{x=60.321926174000055, y=169.73389733200008},
{x=60.52765534100007, y=170.98365319100003},
{x=60.94472890800006, y=172.1806746750001},
{x=61.39252350500004, y=173.0839949880001},
{x=61.96002838700008, y=174.61793053500003},
{x=62.81708405200004, y=176.993011915},
{x=62.38027578300006, y=179.13493899800005},
{x=63.32135651200008, y=179.00554446700005},
{x=63.60834381700005, y=178.73308353000004},
{x=64.25421784100007, y=177.813324415},
{x=64.76748281500005, y=176.316742384},
{x=64.77435944200005, y=177.02564537900003},
{x=64.82444896000004, y=177.410492384},

{x=71.23700592700004, y=178.90544681100005, disconnect=true},
{x=69.35594310100004, y=178.8623153000001},
{x=69.88145348200004, y=175.71554179300006},
{x=70.07697174700007, y=170.55933678500003},
{x=69.02362702000005, y=170.9404403000001},
{x=69.55829498900005, y=168.24097741000003},
{x=69.63336823100008, y=162.172211134},
{x=68.67279694200005, y=161.1709090500001},
{x=69.78754303600005, y=159.88697350400003},
{x=71.03432851800005, y=155.51905358200008, name="END_RUSSIA1"},

{x=-83.85637786299992, y=173.29127037900003, name="START_ANTARCTICA", disconnect=true},
{x=-83.36484140399995, y=169.41431725400003},
{x=-82.75652434699992, y=166.4648543630001},
{x=-81.89316171699994, y=162.97730553500003},
{x=-81.1897111959999, y=160.8144637380001},
{x=-80.4704729149999, y=160.6240340500001},
{x=-79.54827239399992, y=160.21656334700003},
{x=-78.69198984199994, y=161.46314537900003},
{x=-78.74537525799991, y=163.55681399800005},
{x=-78.19345468499995, y=165.23454837300005},
{x=-77.29013437299994, y=163.5127059250001},
{x=-76.30934010199991, y=162.70590254},
{x=-75.71257903399993, y=162.46892337300005},
{x=-74.9436174459999, y=162.489593946},
{x=-74.62444426899992, y=163.78728274800005},
{x=-73.99871184699992, y=164.596039259},
{x=-73.57480234199994, y=165.19841556100005},
{x=-73.49732838299991, y=166.79216556100005},
{x=-72.94768645599993, y=166.35621178500003},
{x=-73.19394296699994, y=168.8877059250001},
{x=-72.55201588299991, y=169.72730553500003},
{x=-71.93938567499993, y=170.22966556100005},
{x=-71.66887786299992, y=170.22877037900003},
{x=-71.14560312299994, y=167.9394637380001},
{x=-70.58896249799994, y=166.40984134200005},
{x=-70.6526018209999, y=163.52719160200002},
{x=-70.92896900799991, y=162.27361087300005},
{x=-70.24130624799994, y=160.98585045700008},
{x=-69.31633879999993, y=158.81714928500003},
{x=-69.17685312299994, y=156.26490319100003},
{x=-68.6462541649999, y=154.69711347700002},
{x=-68.76572030999995, y=152.62134850400003},
{x=-68.48691171699994, y=149.78492272200003},
{x=-68.31658294099992, y=146.96241295700008},
{x=-67.5739885399999, y=145.79704837300005},
{x=-67.3928361959999, y=144.14633222700002},
{x=-66.95419687299994, y=142.30241946700005},
{x=-66.5334611959999, y=138.37810306100005},
{x=-66.19003671699994, y=135.56812584700003},
{x=-66.1506486959999, y=132.47828209700003},
{x=-67.08521900799991, y=129.21371504},
{x=-66.8318010399999, y=126.66570071700005},
{x=-66.51702239399992, y=125.32007897200003},
{x=-66.81536223799992, y=122.49293053500003},
{x=-66.97771575299993, y=120.44597415500004},
{x=-67.13836028399993, y=116.30941816500001},
{x=-66.70566171699994, y=115.70826256600003},
{x=-66.14804452899995, y=110.72242272200003},
{x=-66.81804778399993, y=108.62191816500007},
{x=-66.0022111959999, y=103.55990644600007},
{x=-66.49521249799994, y=100.08545983200003},
{x=-66.47356536299992, y=98.55355879000007},
{x=-66.54314544099992, y=94.54509524800005},
{x=-66.77019622199992, y=90.08757571700005},
{x=-67.12607187299994, y=85.09489993600005},
{x=-67.88201262799993, y=80.79493248800003},
{x=-69.24358489399992, y=76.96778405000003},
{x=-69.88616301899992, y=75.35132897200003},
{x=-70.01458098799992, y=73.05616295700008},
{x=-70.89055754999993, y=71.71436608200008},
{x=-71.85637786299992, y=70.43287194100003},
{x=-72.6740861959999, y=67.94800866000008},
{x=-72.65943775799991, y=67.09131920700008},
{x=-71.40048593499995, y=68.03101647200003},
{x=-70.32333749799994, y=67.65788821700005},
{x=-69.37363046699994, y=68.81120853000004},
{x=-68.28525155999995, y=69.81364993600005},
{x=-67.55575937299994, y=61.18572024800005},
{x=-67.27483489399992, y=58.722992384000065},
{x=-67.05282968499995, y=56.28956139400003},
{x=-66.7268205709999, y=57.00147545700008},
{x=-66.57048919099992, y=50.19361412900008},
{x=-67.11532968499995, y=50.14771569100009},
{x=-67.33806731599992, y=49.198008660000085},
{x=-67.6115861959999, y=47.167979363000086},
{x=-67.80331796699994, y=45.68864993600005},
{x=-68.53704192499993, y=41.27198326900009},
{x=-69.62900155999995, y=39.56104576900009},
{x=-69.7146135399999, y=37.54997806100005},
{x=-69.11142343499995, y=34.650726759000065},
{x=-69.75107187299994, y=32.89763431100005},
{x=-70.50017669099992, y=29.163584832000083},
{x=-71.03240325299993, y=26.445648634000065},
{x=-70.7776018209999, y=22.761729363000086},
{x=-70.90536874799994, y=20.228526238000086},
{x=-70.4939104149999, y=17.751149936000047},
{x=-70.29045989399992, y=14.34937584700009},
{x=-70.74041106599992, y=11.321055535000085},
{x=-70.47234465899993, y=7.930430535000085},
{x=-70.59107838299991, y=6.357269727000073},
{x=-70.87249114399992, y=3.187998894000089},
{x=-71.34002044099992, y=-1.32453365799995},
{x=-71.36516692499993, y=-3.75454667899993},
{x=-71.20134856599992, y=-7.559315558999913},
{x=-71.1853166649999, y=-9.105051235999952},
{x=-72.00497812299994, y=-11.036203579999949},
{x=-72.84628671699994, y=-13.491037563999953},
{x=-73.71379973799992, y=-16.494699673999946},
{x=-74.43084075299993, y=-15.639027472999942},
{x=-75.22389088299991, y=-18.749745245999918},
{x=-75.68482838299991, y=-22.622425910999937},
{x=-77.12753671699994, y=-31.88182532499991},
{x=-78.2365861959999, y=-36.03270423099991},
{x=-79.65271032799991, y=-28.69211449599993},
{x=-80.48841725049994, y=-32.48100745349993},
{x=-81.22616952899995, y=-41.11949622299994},
{x=-82.5407854149999, y=-46.042225714999915},
{x=-82.19402434699992, y=-53.79963131399995},
{x=-83.37444426899992, y=-61.52424068899995},
{x=-82.20300329666661, y=-61.16348222533325},
{x=-81.78004322699991, y=-63.479400193999936},
{x=-80.97714609199994, y=-68.44871985599991},
{x=-80.86345794099992, y=-75.14883378799993},
{x=-79.99509856599992, y=-78.78001868419994},
{x=-78.89470794099992, y=-82.56663977799991},
{x=-78.5457333315999, y=-81.9825414701999},
{x=-78.15260182049994, y=-74.47380530549992},
{x=-76.82659270599993, y=-77.18687903599994},
{x=-76.67457447699991, y=-72.89476477799991},
{x=-75.78728606599992, y=-65.01085364499994},
{x=-74.7946916649999, y=-62.60773678299995},
{x=-74.25123463299991, y=-61.75694739499994},
{x=-73.48105234199994, y=-61.48932857999995},
{x=-73.22730885199991, y=-60.46934973899994},
{x=-72.64576588299991, y=-61.42027747299994},
{x=-72.02825286299992, y=-62.13312740799995},
{x=-71.44736093499995, y=-61.5540258449999},
{x=-70.61882903399993, y=-61.68114173099991},
{x=-69.65504322699991, y=-62.38182532499991},
{x=-68.45720794099992, y=-63.11986243399991},
{x=-68.36785247199992, y=-65.06114661399994},
{x=-67.16961028399993, y=-64.85309811099995},
{x=-66.63160572699991, y=-64.08653723899994},
{x=-66.56698984199994, y=-62.84544837099992},
{x=-66.20045338299991, y=-61.73127193899995},
{x=-65.91285572699991, y=-61.24461829299992},
{x=-65.44817473799992, y=-62.097767706999946},
{x=-64.56609465899993, y=-60.26195227799991},
{x=-64.28240325299993, y=-58.76992753799993},
{x=-63.541680596999925, y=-57.38609778599994},
{x=-63.402601820999905, y=-58.10342363199993},
{x=-63.90715911299992, y=-59.93700110599991},
{x=-64.41228606599992, y=-61.496164516999954},
{x=-64.64560312299994, y=-62.36538652299993},
{x=-65.12542083099993, y=-62.99201412699995},
{x=-65.59905364399992, y=-63.88048255099994},
{x=-66.06658294099992, y=-65.24986731699994},
{x=-66.59140390399995, y=-66.09190833199995},
{x=-66.92782968499995, y=-67.10399329299992},
{x=-67.73064544099992, y=-66.73790442599994},
{x=-68.68808359199994, y=-66.99950110599991},
{x=-69.34091562299994, y=-68.29344641799992},
{x=-70.71998463299991, y=-67.66759192599994},
{x=-72.27752044099992, y=-66.92804928299995},
{x=-73.48406340899993, y=-73.15819251199991},
{x=-73.46249765399995, y=-77.07921301999994},
{x=-73.26311614399992, y=-80.81017005099994},
{x=-73.56943124799994, y=-84.46690833199995},
{x=-73.22039153399993, y=-88.26305091099994},
{x=-73.20256926899992, y=-91.92491614499994},
{x=-72.99553801899992, y=-99.8535863919999},
{x=-73.3044572899999, y=-101.85659745999993},
{x=-73.9021135399999, y=-102.89639238199993},
{x=-74.73186614399992, y=-100.01060950399994},
{x=-75.25554778399993, y=-101.62230383999992},
{x=-75.22031015399995, y=-109.99425208199995},
{x=-74.2634416649999, y=-111.00625566299993},
{x=-75.00383879999993, y=-114.26968339799993},
{x=-74.17302825299993, y=-115.12376868399991},
{x=-74.56373463299991, y=-118.67959550699993},
{x=-74.75253671699994, y=-125.93358313699991},
{x=-74.73626067499993, y=-134.5862524079999},
{x=-75.22714609199994, y=-139.7366430329999},
{x=-75.80396900799991, y=-144.17731686099992},
{x=-76.46550872199992, y=-148.19542395699992},
{x=-77.32878997199992, y=-145.9003800119999},
{x=-77.49578215899993, y=-148.0186661449999},
{x=-77.48015715899993, y=-152.93464107999992},
{x=-77.17473723799992, y=-158.22435462099992},
{x=-78.21412525799991, y=-154.75951087099992},
{x=-79.2014299459999, y=-152.10700436099992},
{x=-80.00302499799994, y=-147.87901770699992},
{x=-80.70989348749993, y=-149.17343095599992},
{x=-82.37411874799994, y=-152.89679928299992},
{x=-83.50644296699994, y=-164.11396236899992},
{x=-83.26366952919992, y=-171.2793757803999},
{x=-84.2712541649999, y=-164.1159561839999},
{x=-85.12607187299994, y=-155.50397701699993},
{x=-84.49749114399992, y=-175.24889075399992},

}